/*      @(#)dial.c	1.16     */
/*LINTLIBRARY*/
/***************************************************************
 *      dial() returns an fd for an open tty-line connected to the
 *      specified remote.  The caller should trap all ways to
 *      terminate, and call undial(). This will release the `lock'
 *      file and return the outgoing line to the system.  This routine
 *      would prefer that the calling routine not use the `alarm()'
 *      system call, nor issue a `signal(SIGALRM, xxx)' call.
 *      If you must, then please save and restore the alarm times.
 *      The sleep() library routine is ok, though.
 *
 *	#include <sys/types.h>
 *	#include <sys/stat.h>
 *      #include "dial.h"
 *
 *      int dial(call);
 *      CALL call;
 *
 *      void undial(rlfd);
 *      int rlfd;
 *
 *      rlfd is the "remote-lne file descriptor" returned from dial.
 *
 *      The CALL structure as (defined in dial.h):
 *
 *      typedef struct {
 *              struct termio *attr;    ptr to term attribute structure
 *              int     baud;           transmission baud-rate
 *              int     speed;          212A modem: low=300, high=1200
 *              char    *line;          device name for out-going line
 *              char    *telno;         ptr to tel-no digit string
 *		char 	*device		Will hold the name of the device
 *					used to makes a connection.
 *		int	dev_len		This is the length of the device
 *					used to makes a connection.
 *      } CALL;
 *
 *      The error returns from dial are negative, in the range -1
 *      to -12, (-31 added) and their meanings are:
 *
 *              INTRPT   -1: interrupt occured
 *              D_HUNG   -2: dialer hung (no return from write)
 *              NO_ANS   -3: no answer within 20 seconds
 *              ILL_BD   -4: illegal baud-rate
 *              A_PROB   -5: acu problem (open() failure)
 *              L_PROB   -6: line problem (open() failure)
 *              NO_Ldv   -7: can't open L-devs file
 *              DV_NT_A  -8: specified device not available
 *              DV_NT_K  -9: specified device not known
 *              NO_BD_A -10: no device available at requested baud-rate
 *              NO_BD_K -11: no device known at requested baud-rate
 *		DV_NT_E -12: requested speed does not match
 *
 *      Setting attributes in the termio structure indicated in
 *      the `attr' field of the CALL structure before passing the
 *      structure to dial(), will cause those attributes to be set
 *      before the connection is made.  This can be important for
 *      some attributes such as parity and baud.
 *
 *      As a device-lockout semaphore mechanism, we create an entry,
 *      in the directory #defined as LOCK, whose name is LCK..dev
 *      where dev is the device name taken from the "line" column
 *      in the file #defined as LDEVS.  Be sure to trap every possible
 *      way out of execution in order to "release" the device.
 *      This entry is `touched' every hour in order to keep uucp
 *      from removing it on its 90 minute rounds.
 *      Also, have the system start-up procedure clean all such
 *      entries from the LOCK directory.
 *
 *      With an error return (negative value), there will not be
 *      any `lock-file' entry, so no need to call undial().
 ***************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <termio.h>
#include <errno.h>
#include <dial.h>
#include <setjmp.h>

#define OK		0
#define TIMEOUT		1

#define YES     1               /* mnemonic */
#define NO      0               /* mnemonic */
#define EQUAL  !strcmp         	/* mnemonic */
#define DIFFER  strcmp         	/* mnemonic */
#define say     (void)fprintf   /* mnemonic */

extern unsigned
        sleep(),
        alarm();

extern int errno,
	close();

extern char *malloc();

static char
	cul[15+sizeof(DEVDIR)] = DEVDIR,	/* line's device-name */
	acu_name[20],							/* acu's name */
	*find_dev(),							/* local function */
	lock[16+sizeof(LOCK)];				/* directory in which to make lockfile*/

static int
	sperfg=0,		/* requested speed not available */
	found=0,			/* set when device is seen legal */
	saverr,			/* hide errno during other calls */
	rlfd,				/* fd returned by dial for remote comm line */
	FD,				/* fd used localy for remote comm line */
	lfd= -1,			/* fd for the device-lock file */
	intflag=NO,		/* interrupt indicator */
	(*acu_end)() = close,	/* comm line close routine */
	connect(),		/* local function */
	intcatch(),		/* interrupt routine */
	hupcatch();		/* interrupt routine */

int alrmcatch();	/* interrupt routine */

#ifdef ddt
static void dump();
#endif

char 	device[34];
int    	_debug=0;
void   	undial();
static void	disconnect(),
	unlock();


int
dial(call)
CALL call;
{
        FILE *Ldevices;         /* file pointer for Device name file */

        char dvc[30];
        int (*savint)(), (*savhup)();

#ifdef ddt
        if(_debug == YES) {
                say(stderr, "call dial(%d)\r\n", call);
                dump(&call, 0);
        }
#endif

        saverr = 0;
        savint = signal(SIGINT, intcatch);
        savhup = signal(SIGHUP, hupcatch);
        (void)signal(SIGALRM, alrmcatch);

        if(call.telno == NULL && call.line == NULL) {
                	rlfd = DV_NT_K;
                	goto OUT;
        }

        if((Ldevices = fopen(LDEVS, "r")) == NULL) {
                saverr = errno;
                rlfd = NO_Ldv;
                goto OUT;
        }

        while(1) {
                int xx;

                (void)strcpy(dvc, find_dev(Ldevices, &call));
                if(strlen(dvc) == 0) 
                        goto F1;        /* failure to find device */
                (void)strcpy(lock, LOCK);
                (void)strcat(lock, dvc);

                /* creat will always succeed if usr-id is 0,
                   so  check that case separately */
                if(geteuid() == 0  && access(lock, 0) == 0)  
                        goto F0;        /* device found but busy */ 
                if((lfd = creat(lock, 0444)) < 0) 
                        goto F0;        /* device found but busy */ 

                xx = getpid();
                (void)write(lfd, (char*)&xx, sizeof(xx));
		break; /* we have a device get out of here */
        F0:
                if(!call.line)            /* dial device is busy */
                        continue;       /* try to find another */

        F1:
                if(call.line)
                        if(found)       /* specific device request */
                                rlfd = DV_NT_A;
                        else if(sperfg == 1)  
				rlfd = DV_NT_E;
			else
                                rlfd = DV_NT_K;
		else
                        if(found)        /* we are dialing */
                                rlfd = NO_BD_A;
                        else
                                rlfd = NO_BD_K;
                goto CLOUT;
        }
        if (intflag == YES) 
                rlfd = INTRPT;
        else if ((rlfd = connect(&call)) < 0) 
		unlock();
	else 
		(void)alarm(UPDTE);
CLOUT:
        (void)fclose(Ldevices);
OUT:
        (void)signal(SIGINT, savint);
        (void)signal(SIGHUP, savhup);
        errno = saverr;
        return(rlfd);
}

/***************************************************************
 *      connect: establish dial-out or direct connection.
 *      Negative values returned (-1...-7) are error message indices.
 ***************************************************************/

static int
connect(call)
CALL *call;
{
	int alarmtr();
	int reset_alarm();
        int err = 0, dum, t, w, x;
        unsigned u;
        char *p, sp_code, b[30];
	acu_t *acu;
        struct termio *lvp, lv;

#ifdef ddt
        if(_debug == YES) {
                say(stderr, "call connect(%o)\n", call);
                dump(call, 0);
        }
#endif

        switch(call->baud) {
		case 110:
			sp_code = (B110 | CSTOPB);
			break;
		case 134:
			sp_code = B134;
			break;
		case 150:
			sp_code = B150;
			break;
		case 300:
			sp_code = B300;
			break;
		case 600:
                        sp_code = B600;
                        break;
                case 1200:
                        sp_code = B1200;
                        break;
                case 2400:
                        sp_code = B2400;
                        break;
                case 4800:
                        sp_code = B4800;
                        break;
                case 9600:
                        sp_code = B9600;
                        break;
                default:
                        err = ILL_BD;
                        goto RTN;
        }
#ifdef ddt
	say(stderr, "opening <%s>\n", cul);
#endif
	/*
	 * open the line non-blocking to make sure it will be enabled
	 */
        if((FD = open(cul, O_EXCL | O_RDWR | O_NDELAY)) < 0) {
                perror(cul);
                err = L_PROB;
                goto RTN;
        }
        (void)fcntl(FD, F_SETFL, fcntl(FD, F_GETFL, 0) & ~O_NDELAY);

	if(call->device && call->dev_len !=0) {
		strncpy(call->device, cul, call->dev_len);
		if(strlen(cul) >= call->dev_len)
			call->device[call->dev_len -1] = '\0';
	}
        if(!call->attr) {
		/* if not given a termio struct grab the existing one */
                lvp = &lv;
        	if((t = ioctl(FD, TCGETA, lvp)) < 0) {
                	perror("get tty for remote");
                	err = L_PROB;
                	goto RTN;
        	}
	}
        else
                lvp = call->attr;

        lvp->c_cflag |= (CREAD | HUPCL);
        if(!(lvp->c_cflag & CSIZE))
                lvp->c_cflag |= CS8;
	/*
	 * mark the line as direct until communication is
	 * established with the modem
	 */
	lvp->c_cflag |= CLOCAL;
	lvp->c_cflag &= ~CBAUD;
	lvp->c_cflag |= sp_code;
	lvp->c_lflag &= ~ECHO;

        if((t = ioctl(FD, TCSETA, lvp)) < 0) {
                perror("stty for remote");
                err = L_PROB;
                goto RTN;
        }
#ifdef  ddt
        if(_debug == YES) say(stderr,"value of cflag = %o\n\r", lvp->c_cflag);
#endif
        if(call->telno) {
                for(acu = acutable; acu->acu_name != NULL; acu++) {
                        if(EQUAL(acu_name, acu->acu_name))
                                break;
                }
		if (acu->acu_name == NULL)
                	err = DV_NT_K;	/* device not known (WATCH IT!) */
		else {
			/* point acu_end at device close func */
                	acu_end = acu->acu_disconnect;	
                	err = (*(acu->acu_connect))(call->telno);
		}
        }
RTN:
        if(intflag == YES)
                err = INTRPT;
#ifdef ddt
        if(_debug == YES)
                say(stderr, "connect ends with err=%d, FD=%d\n", err, FD);
#endif
	if(err < 0) {
		if (FD >= 0) /* we might or might not have a line open */
			disconnect(FD);
		return(err);
	} 
	else {
		/*
		 * if a modem line or modem control needed 
		 * for the line, mark it as so
		 */
	        if((call->telno == NULL ) || (call->modem))
	                lvp->c_cflag &= ~CLOCAL;
		return(FD);
	}
}

/***************************************************************
 *      find_dev: find a device with the wanted characteristics
 *            specified in line and baud arguments.
 *      Return pointer to device name for use in lock-out semaphore.
 *      The variable 'cul' will be set to contain the
 *              complete path name of the corresponding device.
 *      If the L-devices list always contains a '0' entry regradless
 *              of if the line is direct or not since we do not
 *		use true dialers.
 *      The global 'acu_name' is set to the brand of ACU.
 ***************************************************************/

static char*
find_dev(iop, call)
FILE *iop;
CALL *call;
{
        char buf[50], typ[4], temp[15], *b;
        int tspeed;

#ifdef ddt
        if(_debug == YES) {
                say(stderr, "call find_dev(%o)\n", call);
                dump(call, 0);
        }
#endif

        if(call->telno == NULL)
                (void)strcpy(typ, "DIR");
        else
                (void)strcpy(typ, "ACU");

        while(fgets(buf, 50, iop) != NULL) {

                if (strchr("# \t\n", buf[0]) != NULL)
                        continue;
                if(DIFFER(typ, strtok(buf, " \t")))
                        continue;

                (void)strcat(strcpy(cul,DEVDIR),strtok((char*)0," \t"));

		/* ignore this field; it is not used */
		b = strtok((char*)0, " \t");

                tspeed = atoi(strtok((char*)0," \t\n"));

                if((b = strtok((char *)0, " \t\n")) != NULL)
                        (void) strcpy(acu_name, b);
                else
                        acu_name[0] = '\0';

                if(call->line) {
                        if(strchr((b=call->line), '/') == 0) {
                                (void)strcpy(temp, DEVDIR);
                                b = strcat(temp, call->line);
                        }
                        if(DIFFER(b, cul))
                                continue;
			if(call->baud < 0) {
				/*found line, no baud rate requested, set */
                        	call->baud = call->speed = tspeed;
			} else if(tspeed != call->baud) {
				/* found line at wrong speed, keep looking */
				sperfg = 1;
				continue;
			} else {
				/* found line at correct speed, clear error */
				sperfg = 0;
			}
                }
                if(call->telno) {
                        if(call->speed != tspeed)
                                continue;
                }
                if(call->baud > call->speed)
                        continue;
                ++found;
                return(1+strrchr(cul, '/'));
        }
        return("");
}

/*
 * the way this is set up you can only dial one number at a time
 * (like the original stuff...)
 * so rlfd does'nt make any difference
 */
void
undial(rlfd)
int rlfd;
{
#ifdef ddt
        if(_debug == YES)
                say(stderr, "call undial(%d)\n", rlfd);
#endif
        if(rlfd < 0)
		say(stderr, "Undial: bad line descriptor %d\n", rlfd);
	else
		disconnect(rlfd);
	unlock();
        return;
}

/*
 * rlfd is passed here because acu_end() can also be close()
 */
static void
disconnect(rlfd)
int rlfd;
{
	(void)(*acu_end)(rlfd);
}

static void
unlock() 
{
	if (lfd < 0)
		say(stderr, "Unlock: bad lock descriptor %d\n", lfd);
	else {
                (void)close(lfd);
                lfd = -1;
                if(unlink(lock) < 0)
                        say(stderr, "Can't unlink lock-file\r\n");
#ifdef ddt
                else if(_debug == YES)
                        say(stderr, "Lock-file unlinked\r\n");
#endif
        }
}


static int
alrmcatch()
{
        (void)alarm(UPDTE);
        (void)utime(lock, (struct {long a,b;} *)0);
        (void)signal(SIGALRM, alrmcatch);
}

static int
hupcatch()
{
        undial(rlfd);
}

static int
intcatch()
{
        intflag = YES;
        (void)signal(SIGINT, intcatch);
}

#ifdef ddt
static void
dump(arg, fd)
CALL *arg;
int fd;
{
        struct termio xv;
        int i;

        if(fd > 0) {
                say(stderr, "\r\ndevice status for fd=%d\r\n", fd);
                say(stderr, "F_GETFL=%o\r\n", fcntl(fd, F_GETFL,1));
                if(ioctl(fd, TCGETA, &xv) < 0) {
                        char buf[100];
                        int x=errno;

                        (void)sprintf(buf, "\rtdmp for fd=%d:", fd);
                        errno = x;
                        perror(buf);
                        return;
                }
                say(stderr, "iflag=`%o',", xv.c_iflag);
                say(stderr, "oflag=`%o',", xv.c_oflag);
                say(stderr, "cflag=`%o',", xv.c_cflag);
                say(stderr, "lflag=`%o',", xv.c_lflag);
                say(stderr, "line=`%o'\r\n", xv.c_line);
                say(stderr, "cc[0]=`%o',", xv.c_cc[0]);
                for(i=1; i<8; ++i)
                        say(stderr, "[%d]=`%o',", i, xv.c_cc[i]);
                say(stderr, "\r\n");
        }
        say(stderr,"baud=%d, ",arg->baud);
        say(stderr,"speed=%d, ",arg->speed);
        say(stderr,"line=%s, ",arg->line? arg->line: "(null)");
        say(stderr,"telno=%s\r\n",arg->telno? arg->telno: "(null)");
}
#endif
/*
 ************************** acu devices ****************************
 */

int hayes_connect_pulse(), hayes_disconnect(), hayes_abort();
int hayes_connect_tone(), hayes_disconnect(), hayes_abort();
int vadic_connect(), vadic_disconnect(), vadic_abort();

acu_t acutable[] = {
	"anchor", hayes_connect_pulse, hayes_disconnect, hayes_abort,
	"fujitsu", hayes_connect_pulse, hayes_disconnect, hayes_abort,
	"fujitsutone", hayes_connect_tone, hayes_disconnect, hayes_abort,
	"fujitsupulse", hayes_connect_pulse, hayes_disconnect, hayes_abort,
	"hayes", hayes_connect_pulse, hayes_disconnect, hayes_abort,
	"hayestone", hayes_connect_tone, hayes_disconnect, hayes_abort,
	"hayespulse", hayes_connect_pulse, hayes_disconnect, hayes_abort,
	"vadic", vadic_connect, vadic_disconnect, vadic_abort,

	0, 0, 0, 0
};

/*
 * Routines for calling up on a Hayes Smartmodem 1200 
 * or any Hayes-compatible modem 
 * interrupts are caught in dial()
 */

static jmp_buf Sjbuf;

static							/* dick */
hayes_connect_pulse(num)				/* dick */
	register char *num;				/* dick */
{							/* dick */
	return(hayes_connect(num,0));			/* dick */
}							/* dick */

static							/* dick */
hayes_connect_tone(num)					/* dick */
	register char *num;				/* dick */
{							/* dick */
	return(hayes_connect(num,1));			/* dick */
}							/* dick */

static
hayes_connect(num,toneflag)				/* dick */
	register char *num;
	register toneflag;
{
	int i, ok;
	char phone[50];

	/*
	 * make sure the line is valid
	 */
	 if (FD < 0) {
		say(stderr, "%s: bad connect line %d\n", acu_name, FD);
		return (L_PROB);
	}

	/*
	 * Get modems attention and check the write
	 */
#ifdef ddt
	fprintf (stderr, "get modems attention...");
#endif
	if (hawrite("\r", 2) < 0)
		return (A_PROB);
/*	I took this out because it seems redundant and takes time - dick */
/*	hawrite("+++", 0);			/*dick*/ 
/*	sleep(4);	/* wait for modem to get my command */

	for (i = 0; i < 3; i++) {
		hawrite("ATZ\r", 0);
		if ((ok = expect("OK", 5)) == OK)
			break;
		sleep(3);			/* dick */
		hawrite("+++", 0);
		sleep(2);	/* wait for modem to disconnect */
	}
	if (ok != OK)
		return (A_PROB);

	sleep(1);
	hawrite("ATS14=105\r",0);		/*dick*/
	expect ("OK",5);			/*dick*/
	hawrite("ATQ0\r", 0);	/* send result codes to DTE */
	expect ("OK",5);			/*dick*/
	hawrite("ATV1\r", 0);	/* send result codes as strings */
	expect ("OK",5);			/*dick*/
	if (toneflag)				/*dick*/
		strcpy(phone, "ATDT");		/*dick*/
	else					/*dick*/
		strcpy(phone, "ATD");		/*dick*/
	strcat(phone, num);
	strcat(phone, "\r");
	hawrite(phone, 0);

#ifdef ddt
	fprintf (stderr, "\ndialing...");
#endif
	ok = expect("CONNECT", 60);
	if (ok < 0)
		return (A_PROB);
	else if (ok == TIMEOUT)
		return (NO_ANS);
	else
		return (OK);
}

static
hayes_disconnect()
{
	register i, retry;
	int ok;

#ifdef ddt
	fprintf (stderr, "...start disconnecting...");
#endif

	retry = 3;
	do {
		sleep(3);
		hawrite("+++", 0);
/*		sleep(4);				dick */
	} while(expect("OK", 5) && --retry>0);
	hawrite("ATH\r", 0);
	if (expect("OK", 5)) {
		fprintf(stderr, "%s: can't disconnect acu on line %d\r\n", acu_name, FD);
		ok == A_PROB;
	}
	else {
		ok == OK;
		hawrite("ATZ\r", 0);
	}
#ifdef ddt
	fprintf(stderr, "closing line %d\n", FD);
	fflush(stderr);
#endif
	close(FD);
	return (ok);
}

static
hayes_abort()
{
	hayes_disconnect();
}

static
vadic_connect(num)
register char *num;
{
	int ok;

	/* make sure line is valid
	 */
	if (FD < 0) {
		say(stderr, "vadic: bad connect line %d\n", FD);
		return(L_PROB);
	}

	/* get Racal-Vadic VA212 Auto-Dialer attention and check write
	 */
#ifdef ddt
	fprintf(stderr, "get modems attention...");
#endif
	if (hawrite("\r", 2) < 0)
		return(A_PROB);
	else if ((ok = expect("*", 5)) < 0)
		return(A_PROB);

#ifdef ddt
	fprintf(stderr, "\n");
	fflush(stderr);
#endif

	hawrite("O\r", 1);
	expect((char *)0, 5);
	hawrite("1\r", 1);
	expect((char *)0, 5);
	hawrite("3\r", 1);
	expect((char *)0, 5);
	hawrite("\r", 1);
	expect((char *)0, 5);

#ifdef ddt
	fprintf(stderr, "\n");
	fflush(stderr);
#endif

	hawrite("D\r", 1);
	sleep(2);
	if (expect("NUMBER?", 5) != OK)
		return(A_PROB);

#ifdef ddt
	fprintf(stderr, "\n");
	fflush(stderr);
#endif

	hawrite(num, 1);
	hawrite("\r", 1);
	sleep(2);
	if (expect(num, 10) != OK)
		return(A_PROB);
	hawrite("\r", 1);

#ifdef ddt
	fprintf(stderr, "dialing...\n");
	fflush(stderr);
#endif

	/* modem sequence is DIALING...,  ANSWER TONE, ON LINE
	 */
	if ((ok = expect("DIALING...", 10)) != OK)
		return(A_PROB);
	else if ((ok = expect("ANSWER TONE", 10)) != OK)
		return(NO_ANS);
	else if ((ok = expect("ON LINE", 30)) != OK)
		return(NO_ANS);
	else
		return(OK);
}

static
vadic_disconnect()
{

#ifdef ddt
	fprintf(stderr, "...disconnect by closing dial-out line...\n");
#endif

	close(FD);
	return(OK);
}

static
vadic_abort()
{
	vadic_disconnect;
}

static
hawrite(cp, delay)
	register delay;
	register char *cp;
{

	for (; *cp; sleep(delay), cp++) {
		if (write(FD, cp, 1) < 0) {
			perror ("write");
			return (-1);
		}
#ifdef ddt
		write(1, cp, 1);
		if (*cp=='\r')
			write(1,"\n",1);
#endif
	}
}

static
expect(cp, timeout)
	register char *cp;
	int timeout;
/* if cp == (char *)0, read any junks coming from the modem */
{
	char buf[300];
	register char *rp = buf;
	int alarmtr(), online = 0;

	*rp = 0;

	online =  ! (strcmp(cp, "CONNECT") && strcmp(cp, "ON LINE"));

	signal(SIGALRM, alarmtr);
	if (setjmp(Sjbuf))
		if (cp)
			return (TIMEOUT);
		else
			return (OK);
	alarm (timeout);

	while (notin(cp, buf) && rp < buf + sizeof (buf) - 1) {

		/* if we are trying for an actual connection do a number of cute 
		 * checks
		 */
		if (online) {
			if ( ! notin("CARRIER", buf)) {
				fprintf (stderr, "%s: lost carrier on line %d\n", acu_name, FD);
				return (-1);
			}
			else if ( ! notin("FAILED", buf)) {
				fprintf (stderr, "%s: failed call on %d\n", acu_name, FD);
				return (-1);
			}
			else if (( ! notin("TONE!", buf)) || ( ! notin("BUSY!", buf))) {
				fprintf (stderr, "%s: no dial tone or busy on %d\n", acu_name, FD);
				return (-1);
			}
			else if ( ! notin("VOICE!", buf)) {
				fprintf (stderr, "%s: voice detected on %d\n", acu_name, FD);
				return (-1);
			}
		}

		if (read(FD, rp, 1) < 0) {
			perror (acu_name);
			return (-1);
		}
		*rp &= 0177;

#ifdef ddt
		if (*rp < ' ' || *rp==0177) {
			fprintf (stderr, "^%c", *rp^0100);
		} else
			fprintf (stderr, "%c", *rp);
		fflush(stderr);
#endif ddt

		if (*rp) 
			rp++;
		*rp = '\0';
	}
	(void)reset_alarm();
	return (OK);
}

static
alarmtr()
{
	(void)signal(SIGALRM, alrmcatch);
	(void)alarm(UPDTE);
	longjmp(Sjbuf, 1);
}

static
reset_alarm ()
{
	(void)signal(SIGALRM, alrmcatch);
	(void)alarm(UPDTE);
}

static
notin(sh, lg)
	char *sh, *lg;
{
	
	if (sh==(char *)0)
		return(1);

	for (; *lg; lg++)
		if (prefix(sh, lg))
			return (0);
	return (1);
}

static
prefix(s1, s2)
	register char *s1, *s2;
{
	register char c;

	while ((c = *s1++) == *s2++)
		if (c == '\0')
			return (1);
	return (c == '\0');
}
