
/*
   #title get_terminal_name -- get a terminal definition file name
*/
/*----------------------------------------------------------------------*
 *                                                                      *
 *         P R O P R I E T A R Y  AND  C O N F I D E N T I A L          *
 *                                                                      *
 *                                                                      *
 *       HARRIS COMPUTER SYSTEMS DIVISION OF HARRIS CORPORATION         *
 *                                                                      *
 *    The software contained herein is furnished pursuant a license     *
 *    agreement for use with a single HCSD computer and cannot be       *
 *    copied except as provided in the license agreement.               *
 *                                                                      *
 *----------------------------------------------------------------------*/

#ifndef VOS
#include <stdio.h>
#define TRUE 1
#define FALSE 0
#endif


/*******************************************************************************
*                                                                              *
*  get_terminal_name - determine the terminfo definition file name             *
*                                                                              *
*******************************************************************************/

/*
 * This routine is called to get the terminal name from the TERM
 * environment variable (or VOS #TER JCL register) and the terminfo
 * directory name from the TERMINFO environment variable (or the VOS
 * #TIN JCL register).  It will ask for the terminal name if it can't
 * figure it out, and it will choose a suitable standard default for
 * the directory if the user has not chosen one.  It uses stderr and
 * stdin (it opens the terminal for input if stdin is not a physical
 * device).
 *
 * For VOS realtime programs, the environment and JCL registers are
 * still checked, even though they probably don't exist.  We use lfn
 * 0 when checking TERMDEF for RT programs only for compatibility,
 * lfn 3 is used for interactive programs.  We don't check if it's a
 * control point.
 *
 * We only ask the user for the terminal type on VOS if this is an
 * interactive type program.  (We don't know what lfns to check if
 * it's not an interactive program.)
 */

int get_terminal_name( file_name )
char *   file_name;

/*
* Parameter definitions:
* ----------------------
*
* file_name  (input/output) char *
*           Name of the terminal or terminfo file.  If a fully-qualified
*           file name, it is not changed.  If it appears to be merely a
*           terminal name, the appropriate directory is prepended.  If
*           NULL, the directory and terminal name are determined somehow.
*/

/******************************  DECLARATIONS  *******************************/
{
#ifdef VOS
#define getenv  Get_env_var
#endif

   extern char * getenv();
#ifdef VOS
   extern char * Get_termdef();
   extern char * Get_jcl_reg();
   extern int    Set_jcl_reg();
   extern int    C_ptyp;     /* program type set by C library */
#endif

   char * dirname;           /* terminfo directory name */
   char * name;              /* terminal type name */
   char dir_buf[512];        /* holding area for directory name */
   char name_buf[512];       /* holding area for terminal type name */

#ifndef USE_NOIO
   FILE * tty_fp;            /* screen file pointer to issue a read on */
   char * p;                 /* for stripping trailing newline */
#endif

/*
 * Define standard default directory name.
 */
#ifdef VOS
#define TIN_DEFAULT     "0000TERI*"
#else
#define TIN_DEFAULT     "/usr/lib/terminfo/"
#endif

/*******************************   CODE BLOCK  *****************************/

   /*
    * Check to see if a fully-qualified name was passed in, or
    * any name at all.
    */
   if ( file_name != NULL && *file_name != '\0' )
   {
      if (
            ( *file_name == '.' && file_name[1] == '/')
         || ( *file_name == '/' )
#ifdef VOS
         || ( strchr(file_name, '*') != NULL )
#endif
         )
         return( TRUE );
      else /* partial name, just hold on to it */
      {
         strcpy( name_buf, file_name );
         name = name_buf;
      }
   }
   else /* null parameter, determine terminal type */
   {
      name = getenv( "TERM" );

#ifdef VOS
      if ( name == NULL || *name == '\0' )
      {
         name = Get_jcl_reg( "TER", name_buf );
         if ( (name == NULL || *name == '\0') && C_ptyp <= 0 )
            name = Get_termdef( (C_ptyp==0 ? 3 : 0), name_buf );
      }
#endif

   /*
    * If no name found, ask the user what kind of terminal this is.
    */
      if ( name == NULL || *name == '\0' )
      {
#ifdef VOS
         if ( C_ptyp != 0 )     /* Give up if not interactive */
#else
         if ( ! isatty(2) )     /* Give up if stderr not a terminal */
#endif
            return( FALSE );
         name = name_buf;

#ifdef USE_NOIO
         IOW (&SCREEN, &SYM_WRITE, &18, " Terminal type?   ", &IO_A_reg);
         IOW (&SCREEN, &SYM_READ, &24, name_buf, &IO_A_reg);
#else
         fputs( "Terminal type? ", stderr );
         fflush( stderr );

         if ( isatty(0) )
            gets( name_buf );
         else
         {
            tty_fp = fopen( ctermid(), "r" );
            if ( tty_fp == NULL )
               return( FALSE );
            p = fgets( name_buf, 24, tty_fp );
            fclose( tty_fp );
            if ( p == NULL )
               return( FALSE );
            for ( p = name_buf; *p && *p != '\n'; ++p )
               ;
            *p = '\0';
#endif
         }

         if ( *name_buf == '\0' || *name_buf == ' ' )
            return( FALSE );

#ifdef VOS
         Set_jcl_reg( "TER", name_buf, 0401 );
#endif
      }
   }

   /*
    * Maybe we are done, if the response was fully qualified.
    */
   if ( *name == '/'
#ifdef VOS
      || strchr(name, '*') != NULL
#endif
      )
   {
      strcpy( file_name, name );
      return( TRUE );
   }

   /*
    * Now get the directory name from the environment.
    */

   dirname = getenv( "TERMINFO" );

#ifdef VOS
   if ( dirname == NULL || *dirname == '\0' )
      dirname = Get_jcl_reg( "TIN", dir_buf );
#endif

   if ( dirname == NULL || *dirname == '\0' )
      strcpy( file_name, TIN_DEFAULT );
   else
      strcpy( file_name, dirname );

   /*
    * Concatenate directory name and terminal name, inserting
    * the appropriate separators.
    */

   p = &(file_name[strlen(file_name) - 1]);
#ifdef VOS
   if ( *file_name == '/' )
   {
      if ( *p != '/' )
         *++p = '/';
      *++p = name[0];
      *++p = '/';
   }
   else if ( *p != '*' )
      *++p = '*';
   getenv(NULL);        /* free heap space */
#else
   if ( *p != '/' )
      *++p = '/';
   *++p = name[0];
   *++p = '/';
#endif

   *++p = '\0';
   strcat( p, name );

#if DEBUG
   printf( "get_terminal_name: teri file name is <%s>\n", file_name );
#endif

   return( TRUE );
}
