/*
**	Title:	QT.C  Quarter inch cartridge Tape driver for unix system V
**
**	Author: John R. Franks
**	Date:	26-Apr-86
**
**	The qt tape driver drives a emulex MT01 scsi tape adapter attached to
**  a cipher 540 1/4" cartridge tape drive.
**
**	Since the tape drive is attached to a SCSI buss; all tape access go
**  through the SCSI controller driver.  Please refer to the SCSI driver (sc.c)
**  for further details of the communications between the two drivers.
**
**	1/8/87:  Minor mods added for MT03/Tandberg.  AYT.
**
** Adapted for HCX
** Author: Terry Treder
** Date:   21-May-87
*/

#include	"../sys/param.h"
#include	"../sys/systm.h"
#include	"../ml/mtpr.h"
#include	"../sys/vm.h"
#include	"../sys/buf.h"
#include	"../ml/pte.h"
#include	"../sys/dir.h"
#include	"../sys/user.h"
#include	"../sys/file.h"
#include	"../sys/uio.h"
#include	"../sys/ioctl.h"
#include	"../sys/mtio.h"
#include	"../sys/errno.h"
#include	"../sys/cmap.h"
#include	"../sys/vbavar.h"
#include	"../sys/scsi.h"
#include	"../sys/qt.h"
#include	"../sys/elog.h"		/* for system activity */
#include	"../sys/iobuf.h"	/* for system activity */
#include	"../sys/kernel.h"	/* for system activity */


#define ROUND_UP(x)	((x % QT_BSIZE) ? (x + (QT_BSIZE - (x % QT_BSIZE))) : x)

/*
** Declarations for ioctl subroutines (needed for jump table below).
*/

extern int	qtwrite_filemark(), qtsearch_fm_forw(), qtsearch_fm_back();
extern int	qt_space_forw(), qt_space_back(), qtrewind();
extern int	qtrewind_tape_unl(), qtdrive_status();
extern int	qtraw_read(), qtraw_write(), qtbuf_read(), qtbuf_write();
extern int	cydgo(), qt_retension(), qt_load(), qt_safe();
extern int  qtwrite_eov();
/*
** Jump table for ioctl functions (used in qtstart).
** Note the the order of the entries in cmd_tbl and the definitons QTC_*
** are tied together. So be careful that no mismatches occur.
*/

static int	(*cmd_tbl[])() = {
	qtwrite_filemark, qtsearch_fm_forw, qtsearch_fm_back, qt_space_forw,
	qt_space_back, qtrewind, qtrewind_tape_unl, qtdrive_status,
	qt_retension, qt_load, qtwrite_eov,
	qtraw_read, qtraw_write, qtbuf_read, qtbuf_write
};

#define	QTC_WFM	0	/* Write file mark */
#define	QTC_SFMF	1	/* Search file mark forward */
#define	QTC_SFMB	2	/* Search file mark backward */
#define	QTC_SPF	3	/* Space forward */
#define	QTC_SPB	4	/* Space backward */
#define	QTC_REW	5	/* Rewind tape */
#define	QTC_UNL	6	/* Rewind and unload */
#define	QTC_STAT	7	/* Get drive status */
#define	QTC_RETN	8	/* Retension tape */                     
#define	QTC_LOAD	9	/* Load tape and set drive mode bits */ 
#define	QTC_SAFE	10	/* rewind tape and turn off safe light */
#define  QTC_RRD  11 /* Raw read */
#define	QTC_RWRT	12	/* Raw write */
#define	QTC_BRD	13	/* Buffered read */
#define	QTC_BWRT	14	/* Buffered write */

#define	TRUE	1
#define	FALSE	0


extern unsigned		qtminsize();

/* Define driver structures for UNIX */

extern	nqt;		/* number of tape drives */
extern	nsc;		/* number of sc devices */

extern	qt_tab		qt_info[];


/* for error logging */

struct qt_reg {
	u_int control;
	u_int status;
} qt_reg;

extern int	blkacty;
extern struct	iotime	qtstat[];	/* for system activity */
extern struct	iobuf	qttab[];	/* for error logging */


/*
**	Open the tape drive for exclusive use of the caller.
** Fails if drive is already open, the scsi controller finds an error
** or the tape fails to load.
*/

qtopen(dev, flags)
register dev_t	dev;
register int	flags;
{
	register int	unit = QTUNIT(dev);
	register qt_tab	*q_info;
	register int	error;

	/* If a unit number is too big then tell them that it does no exist */
	if(unit >= nsc)
		return ENXIO;

	q_info = &qt_info[unit];
	qttab[unit].io_nreg = (sizeof(struct qt_reg) / sizeof(u_int));
	qttab[unit].io_stp =  &qtstat[unit].ios;
	if(q_info->qt_open == FALSE) {
		if(!(error = scopen(unit))) {
			q_info->qt_open = TRUE;
			q_info->qt_ready = TRUE;
			q_info->qt_eof = FALSE;
			q_info->qt_eot = FALSE;
			q_info->open_flags = flags;
			q_info->qt_fm_needed = FALSE;
			q_info->write_protect = FALSE;
			q_info->error_count = 0;

			/*
			**  If the first load returns a hardware error then
			** the tape may have been just loaded.
			** so try it again.
			*/
			if(!(error = qtcmd(dev, QTC_LOAD, 1)))
				return 0;
			scclose(unit);
			q_info->qt_open = FALSE;
		}
		return error;
	}
	return EBUSY;
}


/*
*/

qtclose(dev, flag)
register dev_t dev;
register flag;
{
	register int	unit = QTUNIT(dev);
	register qt_tab	*q_info = &qt_info[unit];
	register int	error;

	/* Make sure drive did not drop offline */
	if(q_info->qt_ready == TRUE) {
		if(q_info->qt_fm_needed) {
			q_info->qt_fm_needed = FALSE;
			qtcmd(dev, QTC_WFM, 1);
		}
		if(!(minor(dev) & NOREWIND))
			error = qtcmd(dev, QTC_UNL, 1);
	}
	scclose(unit);
	if(q_info->error_count != 0) {
		uprintf("\nqt%d: Warning - %d ", unit, q_info->error_count);
		uprintf("tape block%s used for recovering bad spots.\n",
		    (q_info->error_count == 1) ? " was" : "s were");
	}
	q_info->qt_open = FALSE;
}


/*
**	Qtcmd is used internally to implement all the ioctl functions
**  that are needed by the driver.  We duplicate the code in physio
**  that is used for syncronizing the processes (slumber / wakeup) so
**  that we can treat our internal command requests exactly like
**  regular reads and writes.  They get put on the controller queue,
**  start processes them and iodone is called to wake us up on completion.
**
**	We don't call physio directly because it expects data to be moved
**  and has a lot more overhead than we really need.
*/

qtcmd(dev, command, count)
register dev_t	dev;
register long	command;
register int	count;
{
	register int		unit = QTUNIT(dev);
	register qt_tab		*q_info = &qt_info[unit];
	register unsigned 	error;
	register int		priority = spl/**/SCSI_LEVEL();
	
	while (q_info->qtrequest.b_flags & B_BUSY) {
		q_info->qtrequest.b_flags |= B_WANTED;
		slumber(&q_info->qtrequest, PRIBIO, iocomboost);
	}
	splx(priority);

	/* load the request queue element */
	q_info->qtrequest.b_flags &= ~(B_DONE|B_ERROR);
	q_info->qtrequest.b_error = 0;
	q_info->qtrequest.b_dev = dev;
	q_info->qtrequest.av_back = (struct buf *)command;
	q_info->qtrequest.b_bcount = count;
	q_info->qtrequest.b_flags = B_PHYS | B_BUSY;

	scqueue(&q_info->qtrequest, &q_info->qtqueue);
	qtstart(q_info, &q_info->qtrequest);

	q_info->qtrequest.b_flags &= ~(B_PHYS | B_BUSY);
	if(q_info->qtrequest.b_flags & B_WANTED)
		wakeup (&q_info->qtrequest);
	return geterror(&q_info->qtrequest);
}


/*
*/

qtstrategy(request)
register struct buf *request;
{
	register int		unit = QTUNIT(request->b_dev);
	register int		priority;
	
	request->b_start = time.tv_sec;
	qtstat[unit].io_bcnt += ROUND_UP(request->b_bcount) / 512;

   /* The way I read this the guy is saying our prefered buffer size
	 * is QT_BSIZE if the buffer isn't a multiple of QT_BSIZE and
	 * is the buffer size +1 if it is a multiple of QT_BSIZE.
	 *
	 * I think he's trying to force a buffer copy. - tnt
	 */
	buf_setup(request,
	    (request->b_bcount%QT_BSIZE)?QT_BSIZE:(request->b_bcount+1));
	if(request->b_flags & B_PHYS) {
		if(request->b_flags & B_READ)
			request->av_back = (struct buf *)QTC_RRD;
		else
			request->av_back = (struct buf *)QTC_RWRT;
	}
	else
		if(request->b_flags & B_READ)
			request->av_back = (struct buf *)QTC_BRD;
		else
			request->av_back = (struct buf *)QTC_BWRT;
	scqueue(request, &qt_info[unit].qtqueue);
	qtstart(&qt_info[unit], request);
}



/*
**
*/

qtstart(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register struct buf	*que = &q_info->qtqueue;
	register int		unit = QTUNIT(request->b_dev);
	register int		priority = spl/**/SCSI_LEVEL();

	/* Go to slumber until we are at head of queue */
	while(que->av_forw != request) {
		if(slumber(request, PRIBIO-1, iocomboost)) {
			/* get out early if process was killed */
			scremqueue(que, request);
			splx(priority);
			request->b_resid = request->b_bcount;
			request->b_error = EINTR;
			request->b_flags |= B_ERROR;
    			request->b_flags &= ~B_NOT1K;
			iodone(request);
			return;
		}
	}
	splx(priority);

	/* execute the proper command */
	(*cmd_tbl[(int)request->av_back])(q_info, request);

	/* begin system activity and error logging */
	if (qttab[unit].b_active) {
		if (qttab[unit].io_erec)
			logberr(&qttab[unit], request->b_flags & B_ERROR);
		qttab[unit].b_errcnt = 0;
		qttab[unit].b_active = 0;
		qttab[unit].b_actf = request->av_forw;
		qtstat[unit].io_resp += time.tv_sec - request->b_start;
		qtstat[unit].io_act += time.tv_sec - qttab[unit].io_start;
		qttab[unit].io_start = 0;
	}
	q_info->last_resid = request->b_resid;
	scdequeue(que);
    	request->b_flags &= ~B_NOT1K;
	iodone(request);
	if(que->av_forw != NULL)
		wakeup(que->av_forw);
}


/*
**	qtraw_read handles the read requests from the raw device (qtread).
**
**  The special cases are:
**  1)	we can not read after a write.  (writting defines end of file)
**  2)  reading past end of file returns 0 bytes;
*/

qtraw_read(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register int	unit = QTUNIT(request->b_dev);
	register int	addr, command, bytes;

	if(q_info->qt_fm_needed == TRUE) {
		request->b_resid = request->b_bcount;
		request->b_error = ENXIO, request->b_flags |= B_ERROR;
		return;
	}

	addr = get_ioadr(request,q_info->qtraw,q_info->map,q_info->utl);
	qtexec(QT_READ,
	    QT_RW | (ROUND_UP(request->b_bcount) / QT_BSIZE), request, addr);
	q_info->blknum += (request->b_bcount - request->b_resid) / QT_BSIZE;
	end_transfer(request, q_info->qtraw, q_info->map, q_info->utl);
	if(q_info->qt_eof) {
		q_info->qt_eof = 0;
		request->b_resid = request->b_bcount;
		return;
	}
}


/*
**	qtraw_write handles the write requests from the raw device.
**
**  The special cases are
**  1) we don't allow writes after end of tape is reached.
*/

qtraw_write(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register int	unit = QTUNIT(request->b_dev);
	register int	addr, clear_count;

	if(q_info->qt_eot) {
		request->b_resid = request->b_bcount;
		request->b_error = ENXIO, request->b_flags |= B_ERROR;
		return;
	}
	q_info->qt_fm_needed = TRUE;
	addr = get_ioadr(request,q_info->qtraw,q_info->map,q_info->utl);
	if(clear_count = (request->b_bcount % QT_BSIZE)) {
		register char *start=((char *)q_info->qtraw)+request->b_bcount;

		bzero(start, clear_count);
	}
	qtexec(QT_WRITE,
	    QT_RW | (ROUND_UP(request->b_bcount) / QT_BSIZE), request, addr);
	q_info->blknum += (request->b_bcount - request->b_resid) / QT_BSIZE;
	end_transfer(request, q_info->qtraw, q_info->map, q_info->utl);
}


/*
**	qtwrite_filemark processes the ioctl to write filemarks to tape.
*/

qtwrite_filemark(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register int		unit = QTUNIT(request->b_dev);

	/* flush block device buffer first */
	qtbuf_flush(q_info, request);
	qtexec(QT_WRITE_FM, request->b_bcount, request, 0);
	if(request->b_bcount) {
		q_info->qt_fm_needed = FALSE;
		q_info->filenum += (request->b_bcount - request->b_resid);
		q_info->blknum = 0;
	}
}


/*
**	qtsearch_fm_forw is the ioctl to search for a filemark in the
**  forward direction on tape.
**
**  The special cases are:
**  1) if the last command was a write the we can't search.
*/

qtsearch_fm_forw(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register int		unit = QTUNIT(request->b_dev);

	if((q_info->qt_fm_needed == TRUE) || q_info->qt_eot) {
		request->b_resid = request->b_bcount;
		goto	error;
	}
	if(!request->b_bcount)
		request->b_bcount = 0;
	qtexec(QT_SPACE, (QT_FILES|request->b_bcount), request, 0);
	q_info->filenum += request->b_bcount - request->b_resid;
	q_info->blknum = 0;
	if(request->b_resid) {
error:		request->b_error = ENXIO;
		request->b_flags |= B_ERROR;
	}
}


/*
**	qtsearch_fm_back is the ioctl to search for a filemark in the
**  reverse direction on tape.
**
**	Since only one device can be active on a given controller at any
**  given instant in time, we try to be nice and let onther devices  on
**  this controller be scheduled after we space over each record.  This will
**  at least give the apperance of overlapped operations on the controller.
**
**  The special cases are:
**  1) can't search past begining of tape.
**  2) if the last operation was a write data then we need to add
**     an end of volume record before we start searching.
*/

qtsearch_fm_back(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register int	unit = QTUNIT(request->b_dev);
	register int	new_file;

	if(!request->b_bcount)
		request->b_bcount = 1;
	new_file = q_info->filenum - (request->b_bcount - 1);
	if((q_info->blknum == 0) && (q_info->filenum != 0))
		new_file--;
	qtwrite_eov(q_info, request);
	qtexec(QT_REWIND, 0, request, 0);
	q_info->filenum = 0;
	if(new_file > 0) {
		qtexec(QT_SPACE, QT_FILES|new_file, request, 0);
		q_info->filenum = new_file - request->b_resid;
		if(request->b_resid)
			goto	error;
	}
	else if(new_file < 0) {
		request->b_resid = -new_file;
error:		request->b_error = ENXIO;
		request->b_flags |= B_ERROR;
	}	
	q_info->blknum = 0;
}


/*
**	qt_space_forw is used to search forward a given number of records on
**  tape.
**
**	Since only one device can be active on a given controller at any
**  given instant in time, we try to be nice and let onther devices  on
**  this controller be scheduled after we space over each record.  This will
**  at least give the apperance of overlapped operations on the controller.
**
**  The special cases are:
**  1) we can't space over a filemark.
**  2) if the last command was a write data or filemark we can't space forward.
*/

qt_space_forw(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register int		unit = QTUNIT(request->b_dev);

	if((q_info->qt_fm_needed == TRUE) || q_info->qt_eof) {
		request->b_resid = request->b_bcount;
		request->b_error = ENXIO, request->b_flags |= B_ERROR;
		return;
	}

	if(!request->b_bcount)
		request->b_bcount = 1;
	qtexec(QT_SPACE, QT_BLOCKS|request->b_bcount, request, 0);
	q_info->blknum += request->b_bcount - request->b_resid;
	if(q_info->qt_eof) {
		request->b_resid = request->b_bcount;
		request->b_error = ENXIO, request->b_flags |= B_ERROR;
	}
}


/*
**	Qt_space_back spaces backward a given number of records.
**
**	Since only one device can be active on a given controller at any
**  given instant in time, we try to be nice and let onther devices  on
**  this controller be scheduled after we space over each record.  This will
**  at least give the apperance of overlapped operations on the controller.
**
**  The special cases are:
**  1) we can't space over a filemark.
**  2) we can't space past the beginning of tape.
**  3) if the last operation was a write data then we need to add
**     an end of volume record before we start searching.
*/

qt_space_back(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register int	new_blknum;

	if(!request->b_bcount)
		request->b_bcount = 1;
	new_blknum = q_info->blknum - request->b_bcount;
	if(new_blknum < 0)
		new_blknum = 0;
	qtwrite_eov(q_info, request);
	qtexec(QT_REWIND, 0, request, 0);
	if(q_info->filenum) {
		qtexec(QT_SPACE, QT_FILES|q_info->filenum, request, 0);
		if(request->b_resid) {
			q_info->filenum -= request->b_resid;
			q_info->blknum = 0;
			goto	error;
		}
	}
	if(new_blknum) {
		qtexec(QT_SPACE, QT_BLOCKS|new_blknum, request, 0);
		if(request->b_resid) {
			q_info->blknum = new_blknum - request->b_resid;
error:			request->b_error = EIO;
			request->b_flags |= B_ERROR;
			return;
		}
	}
	q_info->qt_eof = FALSE;
	q_info->blknum = new_blknum;
}


/*
**
*/

qt_load(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register qt_mode	*mode_params = (qt_mode *)q_info->qtraw;
	register int		old_count = request->b_bcount;

	qtexec(QT_TEST, 0, request, 0);
	if(((q_info->last_control == S_VENDOR)&&(q_info->last_status == SC_UA))
	    || ((q_info->filenum == 0) && (q_info->blknum == 0))) {
			request->b_error = 0;
			request->b_flags &= ~B_ERROR;
			bzero(mode_params, sizeof(qt_mode));
			mode_params->bufmode = 1;
			mode_params->disc_len = 8;
			mode_params->density_code = 5;
			request->b_bcount = 16;
			qtexec(QT_MODE_SELECT, 13, request, 0);
			if(!(request->b_flags & B_ERROR))
				qtexec(QT_LOAD, QT_LOADIT, request, 0);
			request->b_bcount = old_count;
			return;
		}
}


/*
**
*/

qt_retension(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	qtexec(QT_LOAD, QT_RET, request, 0);
}



/*
** I believe that the following comment is pure bullshit.
** I think ALL commands are synchronous.
**
**	qtrewind_tape_ta does a waiting rewind to of the tape.
**
**  	An overlapped rewind is issued and then we change the command type to
**  a wait for ready ioctl.  Wait for ready contains the logic to poll
**  without blocking anything in the system, until the drive becomes ready or
**  drops off line whichever comes first.
*/

qtrewind(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	/* for system activity */
	q_info->qtqueue.b_active = SREW;

	qtwrite_eov(q_info, request);
	q_info->qt_eof = FALSE;
	q_info->qt_eot = FALSE;
	q_info->filenum = 0;
	q_info->blknum = 0;
	qtexec(QT_REWIND, 0, request, 0);
	request->b_resid = 0;
}


/*
**	qtrewind_tape_unl does an overlapped rewind and then unloads the
**  tape after the tape completes the rewind.  This feature is handled by
**  the individual tape drive and in some cases can not unload a tape.  In
**  this case it acts exactly like an overlapped rewind.
*/

qtrewind_tape_unl(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	qtrewind(q_info, request);
	qt_safe(q_info, request);
}


/*
**
*/

qt_safe(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	qtexec(QT_LOAD, QT_UNLOADIT, request, 0);
}


/*
**	qtdrive_status is used to process the status ioctl request.
**  it depends entirly on the interupt routines to load the last_XXX
**  registers in qt_info[].
*/

qtdrive_status(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register qt_mode	*mode_params = (qt_mode *)q_info->qtraw;
	register int		old_count = request->b_bcount;
	register int		i;

	request->b_bcount = 16;
	qtexec(QT_MODE_SENSE, 13, request, 0);
	request->b_bcount = old_count;
	for(i=0; i<5; i++) {
		uncache(&q_info->qtraw[i]);
	}
	q_info->write_protect = mode_params->write_protected;
}


/*
**	qtbuf_read handles the read requests from the block device.
**
**  The special cases are:
**  1)	we can not read after a write.  (writting defines end of file)
**  2)  reading past end of file returns 0 bytes;
**  3)  if we are mispositioned we have to seek to the correct block.
**  4)  we can hit end of tape while seeking.
**  5)  we want to be nice to other processes while seeking so we
**  	break the request up into smaller requests.
**  6)  returns error if the block was larger than requested. 
*/

qtbuf_read(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register int		unit = QTUNIT(request->b_dev);
	register int		addr;

	if(q_info->qt_eof) {
		request->b_resid = request->b_bcount;
		return;
	}

	if(qtseek(request, q_info, TRUE)) {
		request->b_flags |= B_ERROR;
		request->b_resid = request->b_bcount;
		request->b_error = EIO;
		return;
	}

	if(q_info->buf_in_use) {
		bcopy(q_info->qtraw, request->b_un.b_addr, request->b_bcount);
		request->b_resid = 0;
		return;
	}

	addr = get_ioadr(request,q_info->qtraw,q_info->map,q_info->utl);
	qtexec(QT_READ, QT_RW | (request->b_bcount / QT_BSIZE), request, addr);
	q_info->blknum += (request->b_bcount - request->b_resid) / QT_BSIZE;
	end_transfer(request, q_info->qtraw, q_info->map, q_info->utl);

	/* 
	**   if file is open for a write and end of data is reached on 
	** block 0 of the file then give
	** the user a block of zeros to chew on (this is for cat by the way)
	** seek takes care of the read after write case.
	*/
	if(q_info->open_flags & FWRITE)
		if((q_info->qt_eof && (q_info->blknum == 0)) ||
		    ((q_info->qt_com.scack.ack_sense.s_class == S_VENDOR) &&
		    (q_info->qt_com.scack.ack_sense.s_code == SC_REOM))) {
			if(q_info->blknum != 0)
				q_info->blknum -= QT_BLK_FACTOR;
			request->b_flags &= ~B_ERROR;
			clrbuf(request);
			request->b_resid = 0;
		}
}


/*
**	qtbuf_write handles the write requests from the block device.
**
**  The special cases are:
**  1)  if we are mispositioned we have to seek to the correct block.
**  2)  we can hit end of tape while seeking.
**  3)  we want to be nice to other processes while seeking so we
**  	break the request up into smaller requests.
**  4) we don't allow writes after end of tape is reached.
*/

qtbuf_write(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register int		unit = QTUNIT(request->b_dev);
	register int		addr;
	
	if(qtseek(request, q_info, FALSE)) {
		request->b_flags |= B_ERROR;
		request->b_resid = request->b_bcount;
		request->b_error = EIO;
		return;
	}
	q_info->qt_fm_needed = TRUE;
	bcopy(request->b_un.b_addr, q_info->qtraw, request->b_bcount);
	request->b_resid = 0;
	q_info->buf_in_use = 1;
}


/*
**	qtseek is used by the block device to position the tape correctly
**  before each read or write request.
**
**  The special cases are:
**  1)  we can hit end of tape while seeking.
*/

qtseek(request, q_info, read)
register struct buf	*request;
register qt_tab		*q_info;
int			read;
{
	register int	unit = QTUNIT(request->b_dev);
	register	req_pos = request->b_blkno * QT_BLK_FACTOR;
	struct buf	tmp;

 	tmp = *request;
	if(req_pos == q_info->blknum)
		return 0;
	if(q_info->blknum > req_pos) {
		qtrewind(q_info, &tmp);
		q_info->blknum = 0;
		if(q_info->filenum) {
			qtexec(QT_SPACE, QT_FILES|q_info->filenum, &tmp, 0);
			if(tmp.b_flags & B_ERROR)
				return B_ERROR;
		}
	}
	if(q_info->blknum < req_pos) {
		if(q_info->qt_eot) {
			request->b_resid = request->b_bcount;
			request->b_error = ENXIO, request->b_flags |= B_ERROR;
			return B_ERROR;
		}
		if(qtbuf_flush(q_info, request))
			return B_ERROR;
		if(req_pos == q_info->blknum)
			return 0;
		qtexec(QT_SPACE, QT_BLOCKS|(req_pos-q_info->blknum), &tmp, 0);
		q_info->blknum += req_pos - q_info->blknum - tmp.b_resid;
		if((q_info->blknum < req_pos) && (read == FALSE))
			while((q_info->blknum < req_pos) && !q_info->qt_eot) {
				tmp.b_flags &= ~B_ERROR;
				tmp.b_resid = 0;
				tmp.b_bcount = QT_BLK_FACTOR * QT_BSIZE;
				bzero(q_info->qtraw, tmp.b_bcount);
				qtexec(QT_WRITE, QT_RW|QT_BLK_FACTOR, &tmp, 0);
				q_info->blknum +=
				    (tmp.b_bcount - tmp.b_resid) / QT_BSIZE;
				if(tmp.b_flags & B_ERROR)
					goto out;
			}
	}
out:	return tmp.b_flags & B_ERROR;
}


/*
**
*/

qtbuf_flush(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	if(q_info->buf_in_use) {
		struct buf	tmp;

 		tmp = *request;
		tmp.b_flags &= ~B_ERROR;
		tmp.b_bcount = DEV_BSIZE;
		tmp.b_resid = 0;
		qtexec(QT_WRITE,QT_RW|QT_BLK_FACTOR, &tmp, 0);
		q_info->blknum += (tmp.b_bcount - tmp.b_resid) / QT_BSIZE;
		q_info->buf_in_use = 0;
		return tmp.b_flags & B_ERROR;
	}
	return 0;
}


/*
**
*/

qtwrite_eov(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	if(q_info->qt_fm_needed == TRUE) {
		qtbuf_flush(q_info, request);
		qtexec(QT_WRITE_FM, 1, request, 0);
		q_info->qt_fm_needed = FALSE;
		q_info->filenum++;
		q_info->blknum = 0;
	}
}



/*
*/

qtexec(cmd, depend, request, addr)
register int		cmd, depend;
register struct buf	*request;
caddr_t			addr;
{
	register int	unit = QTUNIT(request->b_dev);
	register qt_tab	*q_info = &qt_info[unit];
	register sc_com	*Cpb = &q_info->qt_com;

	/* I don't believe that you can have or need more than one on a single
	 * SCSI controller.  So the 4th parameter is always 0, instead of
	 * a slave number.
	 */
	load_sc_cmd(Cpb, cmd, 0, 0, depend);
	/* load the request queue element */
	q_info->qtcommand.b_flags &= ~B_ERROR;
	q_info->qtcommand.b_error = 0;
	q_info->qtcommand.b_dev = QTUNIT(request->b_dev);
	q_info->qtcommand.av_back = (struct buf *)Cpb;
	switch (cmd) {
		case QT_READ :
		case QT_WRITE :
			blkacty |= (1 << QTTAPE);
			qtstat[unit].io_cnt++;
			qttab[unit].b_active = SIO;
			q_info->qtcommand.b_bcount=ROUND_UP(request->b_bcount);
			break;
		case QT_MODE_SELECT :
		case QT_MODE_SENSE :
			qtstat[unit].ios.io_misc++;
			qttab[unit].b_active = SCOM;
			q_info->qtcommand.b_bcount = request->b_bcount;
			break;
		default :
			qtstat[unit].ios.io_misc++;
			qttab[unit].b_active = SCOM;
			q_info->qtcommand.b_bcount = 0;
			break;
	}
	q_info->qtcommand.b_resid = 0;
	q_info->qtcommand.b_un.b_addr = (addr == 0) ?
	    (caddr_t)vtoph(0, (int)q_info->qtraw) : addr;
	scstrategy(&q_info->qtcommand);
	blkacty &= ~(1 << QTTAPE);
	if(q_info->qt_com.scack.ack_sense.s_vadd && !q_info->qtcommand.b_resid){
		request->b_resid = q_info->qt_com.scack.ack_sense.s_info;
		if((cmd == QT_READ) || (cmd == QT_WRITE))
			request->b_resid *= QT_BSIZE;
	}
	else
		request->b_resid = q_info->qtcommand.b_resid;
	qt_decode_error(q_info, request);
}


/*
**
*/

qtprint_err(message, val0, val1, val2)
register char	*message;
register long	val0, val1, val2;
{
	uprintf("qt0: ");
	uprintf(message, val0, val1, val2);
	uprintf("!\n");
}


/*
*/

qt_decode_error(q_info, request)
register qt_tab		*q_info;
register struct buf	*request;
{
	register int	unit = QTUNIT(request->b_dev);
	register int	i;
	register char	*j;

	q_info->last_control = q_info->qt_com.scack.ack_sense.s_class;
	q_info->last_status = q_info->qt_com.scack.ack_sense.s_code;
	if(!(q_info->qtcommand.b_flags & B_ERROR)) {
		switch (q_info->qt_com.scack.ack_sense.s_class) {
		case S_DRIVE :
			switch(q_info->qt_com.scack.ack_sense.s_code) {
			case SC_NS :
				return;
			case SC_HE :
				qtprint_err("Non-recoverable hardware error");
				break;
			case SC_DNR :
				qtprint_err("Drive not ready");
				break;
			case SC_MNL :
				qtprint_err("Media not loaded");
				break;
			case SC_EOT :
				qtprint_err("End of tape");
				break;
			case SC_DT :
				qtprint_err("Drive timeout");
				break;
			default:
				qtprint_err("Unknown drive code '0x%x'",
				     q_info->qt_com.scack.ack_sense.s_code);
				break;
			}
			break;
		case S_TARGET :
			switch(q_info->qt_com.scack.ack_sense.s_code) {
			case SC_UDE :
			case SC_BBF :
				qtprint_err("Uncorrectable data error");
				break;
			case SC_BNF :
				qtprint_err("Block not found");
				break;
			case SC_DTE :
				qtprint_err("DMA timeout");
				break;
			case SC_WP :
				qtprint_err("Drive write protected");
				break;
			case SC_CDC :
				q_info->error_count++;
				return;
			case SC_FMD :
				q_info->qt_eof = 1;
				return;
			case SC_CE :
				qtprint_err("Data compare error");
				break;
			default:
				qtprint_err("Unknown target code '0x%x'",
				     q_info->qt_com.scack.ack_sense.s_code);
				break;
			}
			break;
		case S_SYS :
			switch(q_info->qt_com.scack.ack_sense.s_code) {
			case SC_IC :
				qtprint_err("Unauthorized media or invalid command");
				break;
			default:
				qtprint_err("Unknown system code '0x%x'",
				     q_info->qt_com.scack.ack_sense.s_code);
				break;
			}
			break;
		case S_VENDOR :
			switch(q_info->qt_com.scack.ack_sense.s_code) {
			case SC_UA :
				return;
			case SC_CT :
				qtprint_err("Command timeout");
				break;
			case SC_AE :
				qtprint_err("Can only write at end of media");
				break;
			case SC_REOM :
				break;
			default:
				qtprint_err("Unknown vendor code '0x%x'",
				     q_info->qt_com.scack.ack_sense.s_code);
				break;
			}
			break;
		default :
			qtprint_err("Unknown error class '0x%x'",
			     q_info->qt_com.scack.ack_sense.s_class);
			break;
		}
	}
	request->b_flags |= B_ERROR;
	request->b_error = EIO;
	qt_reg.control = q_info->qt_com.scack.ack_sense.s_class;
	qt_reg.status = q_info->qt_com.scack.ack_sense.s_code;
	qttab[unit].b_dev = (QTTAPE << 8) | 0;
	qttab[unit].b_actf = request;
	qttab[unit].io_addr = request->b_un.b_addr;
	fmtberr(&qttab[unit], (u_int)0, (u_int *)&qt_reg);
}


/*
**	Raw read interface to unix.  Since all the checking is done at a
**  lower level we don't check anything here and we simply return the results.
*/

qtread(dev, uio)
register dev_t dev;
register struct uio *uio;
{
	register qt_tab	*q_info = &qt_info[QTUNIT(dev)];
	
	return physio(qtstrategy,&q_info->qtrequest,dev,B_READ,qtminsize,uio);
}


/*
**	Raw write interface to unix.  Since all the checking is done at a
**  lower level we don't check anything here and we simply return the results.
*/

qtwrite(dev, uio)
register dev_t		dev;
register struct uio	*uio;
{
	register qt_tab	*q_info = &qt_info[QTUNIT(dev)];

	return physio(qtstrategy,&q_info->qtrequest,dev,B_WRITE,qtminsize,uio);
}


/*
**	Qtioctl is called by UNIX every time an ioctl call is made by a user
**  program.  We don't really do much here except call qt_cmd to process the
**  ioctl command.  We don't decode the ioctl type here because our internal
**  jump table accessed by start is ordered by the ioctl number passes here.
**
**	The only special processing is in the status command.  This is because
**  we actually return various data to the user's program with only that
**  command.
*/

qtioctl(dev, command, data, flag)
register dev_t		dev;
register int		command;
register struct mtop	*data;
register int		flag;
{
	if(command == MTIOCTOP) 
		switch ((unsigned) (data->mt_op)) {

			case MTWEOF:
				return qtcmd(dev, QTC_WFM, data->mt_count);
			case MTFSF:
				return qtcmd(dev, QTC_SFMF, data->mt_count);
			case MTBSF:
				return qtcmd(dev, QTC_SFMB, data->mt_count);
				break;
			case MTFSR:
				return qtcmd(dev, QTC_SPF, data->mt_count);
				break;
			case MTBSR:
				return qtcmd(dev, QTC_SPB, data->mt_count);
				break;
			case MTREW:
				return qtcmd(dev, QTC_REW, data->mt_count);
				break;
			case MTOFFL:
				return qtcmd(dev, QTC_UNL, data->mt_count);
				break;
			case MTNOP:
				return qtcmd(dev, QTC_STAT, data->mt_count);
				break;
			case MTRETEN:
				return qtcmd(dev, QTC_RETN, data->mt_count);
				break;
			default:
				return EIO;
		}
	else if(command == MTIOCGET) {
		register qt_tab	*q_info = &qt_info[QTUNIT(dev)];

		((struct mtget *)data)->mt_type = MT_ISQT;
		((struct mtget *)data)->mt_dsreg = q_info->last_control;
		((struct mtget *)data)->mt_erreg = q_info->last_status;
		((struct mtget *)data)->mt_resid = q_info->last_resid;
		((struct mtget *)data)->mt_fileno = q_info->filenum;
		((struct mtget *)data)->mt_blkno = q_info->blknum;
		(void)qtcmd(dev, QTC_STAT, 1);
		return 0;
	}
	return ENXIO;
}


/*
**	Qtreset is used to unconditionally reset all controllers to
**  their initial state.
*/

qtreset(vba)
register int	vba;
{
}


/*
**	Qtdump is called during a system crash to dump all of main memory.
**  We don't do any special checking here except we will exit early if
**  an i/o error occurs.  The other point is we poll for completion of each
**  command since we don't want to do any special processing, we are
**  the only process running anyway, and possibly the cpu interupt's
**  were not working anyway. (what if sombody stepped on low core?)
*/

qtdump(dev)
register dev_t	dev;
{
}


/*
*/

unsigned
qtminsize(request)
register struct buf	*request;
{
	if(request->b_bcount > (TBUFSIZ * NBPG))
		request->b_bcount = (TBUFSIZ * NBPG);	
}
