/* async.c */

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/file.h"
#include "../sys/proc.h"
#include "../sys/uio.h"
#include "../sys/kernel.h"
#include "../sys/sysinfo.h"

struct areq *global_next_areq;	/* next areq block to allocate */
extern int max_areq;

/*
 * Initialize structures which are used for asynchronous I/O operations.
 */

void
aioinit()
{
	register int	i;

	if (max_areq > 0) {
		big_bzero(&areq_pool[0], sizeof(struct areq)*max_areq);
		for (i = 0; i < max_areq; i++) {
			areq_pool[i].next_areq = &areq_pool[i+1];
		}
		areq_pool[max_areq-1].next_areq = NULL;
		global_next_areq = &areq_pool[0];
	} else {
		global_next_areq = NULL;
	}
}

/*
 * Handle the SIGIO signal.  If the signal did not result from the
 * completion of an asynchronous I/O operation, simply return.
 * Otherwise handle program-level completion of all I/O operations.
 */

void
arw_comp()
{
	register struct areq	*areq;	/* known to be r12 below */
	register struct areq	**tp;	/* known to be r11 below */
	register struct proc	*p = u.u_procp;
	register int		cnt;
	register int		byte_count;

	while (! NO_PENDING_AREQ(p)) {

		areq = p->p_areq_head;
		asm("remque (r12)");

		/*
		 * Invoke lower levels (lowest level first) to perform any
		 * required cleanup operations.
		 */
		for (cnt = areq->comp_rtn_cnt-1; cnt >= 0; --cnt) {
			(*areq->rtn[cnt])(areq);
		}

		/*
		 * Update accounting information.
		 */
		byte_count = areq->req_byte_count - areq->auio.uio_resid;
		u.u_ru.ru_iochr += byte_count;
#ifndef HCXRT
		if (areq->rw == AIO_READ) {
			sysinfo.readch += (unsigned)byte_count;
		} else {
			sysinfo.writech += (unsigned)byte_count;
		}
#endif

		/*
		 * Store the byte count and completion status in the
		 * user-specified buffers.
		 */
		if (areq->byte_count)
			suword(areq->byte_count, byte_count);
		if (areq->IO_status)
			suword(areq->IO_status, areq->driver_status);

		p->p_procmask |= sbittab[areq->id];
		areq->fp->async_count--;

		/*
		 * If user wants to handle the SIGIO signal, then
		 * enqueue the areq block for sendsig to see.
		 */
		if ((u.u_signal[SIGIO] != SIG_DFL) &&
		    (u.u_signal[SIGIO] != SIG_IGN)) {
			tp = &p->p_usrareq_tail;
			asm("insque (r12), *(r11)");
		} else {
			FREE_AREQ(areq);
		}
	}
}

/*
 * Interrupt level I/O completion routine.  This routine is invoked by
 * device drivers upon receiving an interrupt indicating the completion
 * of an asynchronous I/O operation.  This routine is also invoked by
 * device drivers upon successful cancelation of an asynchronous I/O operation.
 */

void
AIO_comp(areq)
	register struct areq	*areq;	/* known to be r12 below */
{
	register struct areq	**tp;	/* known to be r11 below */
	register struct proc	*p = areq->p;

	p->p_compmask |= sbittab[areq->id];
	tp = &p->p_areq_tail;
	asm("insque (r12), *(r11)");
	if ((p->p_wchan == (caddr_t) &p->p_compmask) &&
	    (p->p_waitmask & p->p_compmask)) {
		wakeup(&p->p_compmask);
	}
	psignal(p, SIGIO);	/* handle rest at program level */
}

/*
 * Cancel all of the pending asynchronous I/O requests for the file, and
 * wait the the I/O requests to complete.  This routine is invoked by the
 * process exit logic, close, exec (even not close on exec files), etc.
 * If no file is specified, this service cancels all pending asynchronous
 * I/O requests for all files.
 */

void
AIO_cancel(fpp)
struct file			*fpp;
{
	register struct file	*fp = fpp;
	register struct proc	*p;
	register int		i;
	register int		proc_mask;
	
	/*
	 * Request driver to cancel asynchronous I/O requests, and wait for
	 * the I/O requests to complete.
	 */
	p = u.u_procp;
	proc_mask = p->p_procmask;
	for (;;) {
		i = ffc(proc_mask) - 1;
		if (i == -1) {
			if (fpp && fp->async_count != 0)
				panic("AIO_cancel");
			return;
		}
		proc_mask |= sbittab[i];
		if (!fpp)
			fp = u.u_areq[i]->fp;
		if (fp == u.u_areq[i]->fp) {
			if (! (p->p_compmask & sbittab[i]))
				(*fp->f_ops->fo_acancel)(u.u_areq[i]);
			spl8();
			if (! (p->p_compmask & sbittab[i])) {
				p->p_waitmask = sbittab[i];
				slumber(&p->p_compmask, 0, iocomboost);
			}
			spl0();
			arw_comp();
			proc_mask = p->p_procmask;
		}
	}
}

/*
 * Asynchronous I/O system services:
 */

acancel()
{
	register struct a {
		int		id;
	} *uap = (struct a *)u.u_ap;
	struct file *fp;

	if ((uap->id < 0) || (uap->id >= MAX_AIO_REQ) ||
	    (u.u_procp->p_compmask & sbittab[uap->id])) {
		u.u_error = EINVAL;
		return;
	}

	fp = u.u_areq[uap->id]->fp;
	u.u_error = (*fp->f_ops->fo_acancel)(u.u_areq[uap->id]);
}

await()
{
	register struct a {
		unsigned	nids;
		int		*ids;
	} *uap = (struct a *)u.u_ap;
	register struct	proc *p = u.u_procp;
	register int imask, omask;

	imask = fuword((caddr_t)uap->ids);
	if (u.u_error) {
		return;
	}

	if (uap->nids <= 31) {
		imask &= (sbittab[uap->nids] - 1);
	}

	if (imask == 0) {
		u.u_error = EINVAL;
		return;
	}

	spl8();

	/*
	 * If some of the specified asynchronous I/O operations have
	 * already completed, then return.
	 */
	if (imask & p->p_compmask) {
		goto done;
	}

	/*
	 * Wait for any specified asynchronous I/O operation to complete.
	 */
	p->p_waitmask = imask;
	slumber(&p->p_compmask, PCATCH|INTERRUPTABLE, iocomboost);
	if ((imask & p->p_compmask) == 0) {
		spl0();
		u.u_error = EINTR;
		return;
	}

done:
	spl0();
	omask = imask & p->p_compmask;
	if (imask != omask) {
		if (suword((caddr_t)uap->ids, omask) == -1) {
			u.u_error = EFAULT;
		}
	}
}

aread()
{
	areadwrite(1);
}

awrite()
{
	areadwrite(0);
}

/*
 * Common code for aread and awrite.
 */

areadwrite(readop)
int				readop;	/* asynchronous read operation */
{
	register struct a {
		int		fdes;
		char		*cbuf;
		unsigned	count;
		int		reqid;
		int		*IO_status;
		int		*byte_count;
	} *uap = (struct a *)u.u_ap;
	register struct proc	*p = u.u_procp;
	register struct file	*fp;
	register struct areq	*areq;

	/*
	 * Validate the specified request id.  If none specified, obtain
	 * a valid request id.
	 */
	u.u_r.r_val1 = uap->reqid;
	if (u.u_r.r_val1 == -1) {
		u.u_r.r_val1 = ffs(p->p_procmask) - 1;
		if (u.u_r.r_val1 == -1) {
			u.u_error = EINVAL;
			return;
		}
	} else {
		if ((u.u_r.r_val1 < 0) || (u.u_r.r_val1 >= MAX_AIO_REQ) ||
		    (! (p->p_procmask & sbittab[u.u_r.r_val1]))) {
			u.u_error = EINVAL;
			return;
		}
	}

	/*
	 * Verify access to the file.
	 */
	GETF(fp, uap->fdes);
	if (readop) {
		if ((fp->f_flag&FREAD) == 0) {
			u.u_error = EBADF;
			return;
		}
	} else {
		if ((fp->f_flag&FWRITE) == 0) {
			u.u_error = EBADF;
			return;
		}
	}

	/*
	 * Obtain an areq block.
	 */
	areq = global_next_areq;
	if (areq == NULL) {
		u.u_error = ENOSPC;
		return;
	}
	global_next_areq = areq->next_areq;

	u.u_areq[u.u_r.r_val1] = areq;

#ifndef HCXRT
	sysinfo.sysread++;
#endif
	areq->rw              = readop ? AIO_READ:AIO_WRITE;
	areq->id              = u.u_r.r_val1;
	areq->p               = p;
	areq->IO_status       = uap->IO_status;
	areq->byte_count      = uap->byte_count;
	areq->auio.uio_segflg = UIO_USERSPACE;
	areq->aiov.iov_base   = (caddr_t)uap->cbuf;
	areq->req_byte_count  = areq->auio.uio_resid = areq->aiov.iov_len =
		uap->count;
	areq->fp              = fp;
	areq->auio.uio_offset = fp->f_offset;
	areq->auio.uio_iov    = &areq->aiov;
	areq->auio.uio_iovcnt = 1;
	areq->comp_rtn_cnt    = 0;

	if (areq->aiov.iov_len < 0) {
		u.u_error = EINVAL;
		FREE_AREQ(areq);
		return;
	}

	/*
	 * Store EINPROGRESS in the user-specified status location.  The
	 * actual I/O completion status will be stored by arw_comp().
	 */
	if (areq->IO_status) {
		if (suword(areq->IO_status, EINPROGRESS) == -1) {
			u.u_error = EFAULT;
			FREE_AREQ(areq);
			return;
		}
	}

	/* note: u.u_fmode not set, as the information is available in
	   areq->fp->f_flag, and for asynchronous case who can rely on
	   information in the u area ??? should it be filled in anyhow ??? */

	fp->async_count++;
	p->p_compmask &= cbittab[areq->id];
	p->p_procmask &= cbittab[areq->id];

	if ((fp->f_flag & FASYNCALLOW) != FASYNCALLOW) {
		/*
		 * Asynchronous I/O is not supported.  Emulate it!!!
		 */
		label_t	lqsave;

		u.u_fmode = fp->f_flag;
		areq->driver_status = 0;
		lqsave = u.u_qsave;
		if (setjmp(&u.u_qsave)) {
			if (u.u_error = 0) {
				u.u_error = EINTR;
			}
		} else {
			u.u_error = (*fp->f_ops->fo_rw)
				(fp, readop ? UIO_READ:UIO_WRITE, &areq->auio);
		}
		u.u_qsave = lqsave;
		if (u.u_error) {
			p->p_procmask |= sbittab[areq->id];
			p->p_compmask |= sbittab[areq->id];
			fp->async_count--;
			FREE_AREQ(areq);
			return;
		}
		fp->f_offset += uap->count;
		AIO_comp(areq);
		return;
	}

	/*
	 * Invoke lower-level routine to perform the I/O operation.
	 */
	u.u_error = (*fp->f_ops->fo_asyncrw)(areq);
	if (u.u_error) {
		p->p_procmask |= sbittab[areq->id];
		p->p_compmask |= sbittab[areq->id];
		fp->async_count--;
		FREE_AREQ(areq);
		return;
	}
	fp->f_offset += uap->count;
}
