#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/kernel.h"
#include "../sys/proc.h"
#include "../cf/ipcconfig.h"

#define BSEMGBL 0	/* no shared memory support */

#define NOT_LONGWORD_ALIGNED(x) ((int)(x)&3)

#define BSEMENTRIES		(BSEMGBL+BSEMMAP)
	/* number of entries in table pointed to by p->p_bsemid_ds */

/*
 * Free an entry in a system binary semaphore table:
 * 'inx' is the index into 'p->p_bsemid_ds' for the semaphore
 * 'bsemp' has the value of 'p->p_bsemid_ds[inx]' and is passed as a
 *    parameter for performance reasons
 */
#define BSEMFREE(_inx, _bsemp) \
	{ \
		bsemwake_inter((_bsemp)); \
		if (--((_bsemp)->ref_count) == 0) { \
			(_bsemp)->bsemid_perm.seq++; \
			(_bsemp)->bsemid_perm.mode = 0; \
		} \
		p->p_bsemid_ds[(_inx)] = NULL; \
	}

extern struct bsemid_ds	*ipcget();

/*
 * Initialize the binary semaphore table which resides in local memory.
 */

void
bseminit()
{
	big_bzero(bsemlocal_tbl, sizeof(struct bsemid_ds)*BSEMMAP);
}


/*
 * Convert user supplied bsemid into a ptr to the associated entry in the
 * system binary semaphore table.
 */

struct bsemid_ds *
bsemconv(id)
register int			id;
{
	register struct bsemid_ds	*bsemp;	/* ptr to bsem table entry */

	bsemp = &bsemlocal_tbl[id % BSEMMAP];
	if ((bsemp->bsemid_perm.mode & IPC_ALLOC) == 0 ||
		id / BSEMMAP != bsemp->bsemid_perm.seq) {
		u.u_error = EINVAL;
		return(NULL);
	}

	return(bsemp);
}


/*
 * Allocate process-specific pointers into the system binary semaphore
 * table.  The first BSEMGBL entries in the process-specific table are
 * pointers to binary semaphores in shared memory, and the rest of the
 * entries are pointers to binary semaphores in local memory.  Table
 * entries which are NULL are not in use.
 */

int
bsemallocds(bsemp)
register struct bsemid_ds	*bsemp;	/* ptr to bsem table entry */
{
	register struct proc 		*p = u.u_procp;
	register int			i;

	/*
	 * If this is the first binary semaphore for the process, then
	 * allocate an internal binary semaphore table for it and initialize
	 * the table.
	 */
	if (p->p_bsemid_ds == NULL) {
		p->p_bsemid_ds = (struct bsemid_ds **)
			kmem_alloc(sizeof(int)*BSEMENTRIES);
		if (p->p_bsemid_ds == NULL) {
			u.u_error = ENOMEM;
			bsemp->bsemid_perm.mode = 0;
			return(-1);
		}
		big_bzero(p->p_bsemid_ds, sizeof(int)*BSEMENTRIES);
	}

	/*
	 * Locate an empty entry in the process' binary semaphore table,
	 * and initialize it to contain a pointer to the binary semaphore.
	 */
	for (i = BSEMGBL; i < BSEMENTRIES; i++) {
		if (p->p_bsemid_ds[i] == NULL) {
			p->p_bsemid_ds[i] = bsemp;
			return(0);
		}
	}
	u.u_error = ENOSPC;
	bsemp->bsemid_perm.mode = 0;
	return(-1);
}


/*
 * Wake up all processes waiting on a binary semaphore.
 */

void
bsemwake_inter(bsemp)
register struct bsemid_ds	*bsemp;	/* ptr to bsem table entry */
{
	wakeup(bsemp);
}


/*
 * Handle 'exec' and process exit - free all binary semaphores for the
 * process, and deallocate the process-specific table which is used to
 * point into the system binary semaphore table.
 */

void
bsemexitexec()
{
	register struct proc	 	*p = u.u_procp;
	register struct bsemid_ds	*bsemp;
	register int			i;

	if (p->p_bsemid_ds == NULL) {
		return;
	}

	for (i = BSEMGBL; i < BSEMENTRIES; i++) {
		if ((bsemp = p->p_bsemid_ds[i]) != NULL) {
			BSEMFREE(i, bsemp);
		}
	}

	kmem_free((caddr_t)p->p_bsemid_ds, sizeof(int)*BSEMENTRIES);
	p->p_bsemid_ds = NULL ;
}


/*
 * Binary semaphore system services:
 */

bsemget()
{
	register struct a {
		key_t	key;		/* input */
		struct binsem *binsem;	/* output: all fields */
		int	bsemflg;	/* input */
	} *uap = (struct a *)u.u_ap;
	register struct bsemid_ds	*bsemp;	/* ptr to bsem table entry */
	int				s;	/* ipcget status return */
	int				tmp;	/* temporary variable */

	if ((NOT_LONGWORD_ALIGNED(uap->binsem)) ||
	    (useracc(uap->binsem, sizeof(struct binsem), 0) == 0)) {
		u.u_error = EFAULT;
		return;
	}

	uap->bsemflg |= 0777;	/* diable mode checking in ipcget */

	bsemp = ipcget(uap->key, uap->bsemflg, bsemlocal_tbl, BSEMMAP,
		sizeof(*bsemp), &s);
	if ((bsemp == NULL) || (bsemallocds(bsemp) < 0)) {
		return;
	}

	if (s) {
		/* This is a new binary semaphore.  Finish initialization. */
		bsemp->cpu = 0;
		tmp = 0;
		suword(&uap->binsem->locked, tmp);
		tmp = 1;
		suword(&uap->binsem->not_wanted, tmp);
	}

	bsemp->ref_count++;
	u.u_r.r_val1 = bsemp->bsemid_perm.seq * BSEMMAP +
		(bsemp - bsemlocal_tbl);
}


bsemwait()
{
	register struct a {
		int	bsemid;		/* input */
		struct binsem *binsem;	/* input: locked field */
	} *uap = (struct a *)u.u_ap;
	register struct bsemid_ds	*bsemp;	/* ptr to bsem table entry */

	if ((NOT_LONGWORD_ALIGNED(uap->binsem)) ||
	    (useracc(uap->binsem, sizeof(struct binsem), 0) == 0)) {
		u.u_error = EFAULT;
		return;
	}

	if ((bsemp = bsemconv(uap->bsemid)) == NULL) {
		return;
	}
	if (! uap->binsem->locked) {
		return;
	}

	slumber(bsemp, INTERRUPTABLE, resboost);
}


bsemwakeup()
{
	register struct a {
		int	bsemid;		/* input */
	} *uap = (struct a *)u.u_ap;
	register struct bsemid_ds	*bsemp;	/* ptr to bsem table entry */

	if ((bsemp = bsemconv(uap->bsemid)) == NULL) {
		return;
	}

	bsemwake_inter(bsemp);
}


bsemfree()
{
	register struct a {
		int	bsemid;		/* input */
	} *uap = (struct a *)u.u_ap;
	register struct bsemid_ds	*bsemp;	/* ptr to bsem table entry */
	register struct proc		*p = u.u_procp;
	register int			i;

	if ((bsemp = bsemconv(uap->bsemid)) == NULL) {
		return;
	}

	for (i = 0; i < BSEMENTRIES; i++) {
		if (p->p_bsemid_ds[i] == bsemp) {
			BSEMFREE(i, bsemp);
			return;
		}
	}
}
