/*	kern_clock.c	6.2	83/10/01	*/

#include "../ml/reg.h"
#include "../ml/psl.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/dk.h"
#include "../sys/callout.h"
#include "../sys/user.h"
#include "../sys/kernel.h"
#include "../sys/proc.h"
#include "../sys/vm.h"
#include "../sys/text.h"
#include "../sys/sysinfo.h"

#include "../ml/mtpr.h"
#include "../ml/clock.h"

#ifdef GPROF
#include "../sys/gprof.h"
#endif

#define	PRF_ON	1
extern int prfstat;

/*
 * Clock handling routines.
 *
 * This code is written to operate with two timers which run
 * independently of each other. The main clock, running at hz
 * times per second, is used to do scheduling and timeout calculations.
 * The second timer does resource utilization estimation statistically
 * based on the state of the machine phz times a second. Both functions
 * can be performed by a single clock (ie hz == phz), however the 
 * statistics will be much more prone to errors. Ideally a machine
 * would have separate clocks measuring time spent in user state, system
 * state, interrupt state, and idle state. These clocks would allow a non-
 * approximate measure of resource utilization.
 */

/*
 * TODO:
 *	time of day, system/user timing, timeouts, profiling on separate timers
 *	allocate more timeout table slots when table overflows.
 */

#define BUMPTIME(t, usec) { \
	register struct timeval *tp = (t); \
 \
	tp->tv_usec += (usec); \
	if (tp->tv_usec >= 1000000) { \
		tp->tv_usec -= 1000000; \
		tp->tv_sec++; \
	} \
}

/*
 * The hz hardware interval timer.
 * We update the events relating to real time.
 * If this timer is also being used to gather statistics,
 * we run through the statistics gathering routine as well.
 */
hardclock(pc, ps)
	caddr_t pc;
	int ps;
{
	/* Next 2 dummy variables MUST BE the first local */
	/* variables; leaving place for registers 0 and 1 */
	/* which are not preserved by the 'cct' */

	int	dumm1;		/* register 1 */
	int	dumm0;		/* register 0 */

	register struct proc *p;
	register int s;
	int needsoft = 0;
	extern caddr_t waitloc1, waitloc2;

	if (clk_enable == 0) return;

	/*
	 * Update real-time timeout queue.
	 * At front of queue are some number of events which are ``due''.
	 * The time to these is <= 0 and if negative represents the
	 * number of ticks which have passed since it was supposed to happen.
	 * The rest of the q elements (times > 0) are events yet to happen,
	 * where the time for each is given as a delta from the previous.
	 * Decrementing just the first of these serves to decrement the time
	 * to all events.
	 */
	{
		register struct callout *p1;
		p1 = calltodo.c_next;
		while (p1) {
			if (--p1->c_time > 0)
				break;
			needsoft = 1;
			if (p1->c_time == 0)
				break;
			p1 = p1->c_next;
		}
	}

	/*
	 * Charge the time out based on the mode the cpu is in.
	 * Here again we fudge for the lack of proper interval timers
	 * assuming that the current state has been around at least
	 * one tick.
	 */
	if (USERMODE(ps)) {
#ifndef HCXRT
		if (u.u_prof.pr_scale) {
			addupc(pc, &u.u_prof, 1);
		}
#endif

		/*
		 * CPU was in user state.  Increment
		 * user time counter, and process process-virtual time
		 * interval timer. 
		 */
		BUMPTIME(&u.u_ru.ru_utime, tick);
		if (timerisset(&u.u_timer[ITIMER_VIRTUAL].it_value) &&
		    itimerdecr(&u.u_timer[ITIMER_VIRTUAL], tick) == 0)
			psignal(u.u_procp, SIGVTALRM);
		sysinfo.cpu[CPU_USER]++;
	} else {
		/*
		 * CPU was in system state.
		 */

		if (pc >= waitloc1 && pc <= waitloc2) {
			if (syswait.iowait+syswait.swap+syswait.physio) {
#ifndef HCXRT
				if (syswait.iowait)
					sysinfo.wait[W_IO]++;
				if (syswait.swap)
					sysinfo.wait[W_SWAP]++;
				if (syswait.physio)
					sysinfo.wait[W_PIO]++;
#endif
				sysinfo.cpu[CPU_WAIT]++;
			} else
				sysinfo.cpu[CPU_IDLE]++;
		} else
			sysinfo.cpu[CPU_KERNAL]++;

		if (! noproc) {
			BUMPTIME(&u.u_ru.ru_stime, tick);
		}
	}

	if (noproc == 0) {
		p = u.u_procp;
		/*
		 * Charge the currently running process with resource
		 * utilization for a tick, updating statistics which run
 		 * in (user+system) virtual time, such as the cpu time limit
		 * and profiling timers.
		 * This assumes that the current process has been running
		 * the entire last tick.
		 */
		if ((u.u_ru.ru_utime.tv_sec+u.u_ru.ru_stime.tv_sec+1) >
		    u.u_rlimit[RLIMIT_CPU].rlim_cur) {
			psignal(p, SIGXCPU);
			if (u.u_rlimit[RLIMIT_CPU].rlim_cur <
			    u.u_rlimit[RLIMIT_CPU].rlim_max)
				u.u_rlimit[RLIMIT_CPU].rlim_cur += 5;
		}
		if (timerisset(&u.u_timer[ITIMER_PROF].it_value) &&
		    itimerdecr(&u.u_timer[ITIMER_PROF], tick) == 0)
			psignal(p, SIGPROF);
		s = p->p_rssize;
		u.u_ru.ru_idrss += s;
		if (p->p_textp) {
			register int xrss = p->p_textp->x_rssize;

			s += xrss;
			u.u_ru.ru_ixrss += xrss;
		}
		if (s > u.u_ru.ru_maxrss)
			u.u_ru.ru_maxrss = s;

		/*
		 * In case of a quantum expiration
		 * and reschedule a context switch.
		 */
		p->p_cpticks++;
#ifndef HCXRT
		p->p_cpu++;	/* This is for p_whomp in schedcpu. */
#endif
		if (--quantum == 0) {
			quantum = p->p_qntm;
			if (whichqs != 0) {
				if ((ffs(whichqs)-1) <= p->p_prior) {
					runrun++;
					aston();
				}
			}
		}

	}

	/*
	 * If the alternate clock has not made itself known then
	 * we must gather the statistics.
	 */
	if (phz == 0)
		gatherstats(pc, ps);

	/*
	 * Increment the time-of-day, and schedule
	 * processing of the callouts at a very low cpu priority,
	 * so we don't keep the relatively high clock interrupt
	 * priority any longer than necessary.
	 */
	BUMPTIME(&time, tick);
	if (needsoft) {
#ifdef HCXRT
		p = &proc[3];
		if ((p->p_stat == SSLEEP) && (p->p_wchan == (caddr_t) p)) {
			unsleep(p);
			p->p_stat = SRUN;
			setrq(p);
			if (p->p_prior < u.u_procp->p_prior) {
				runrun++;
				aston();
			}
		}
#else
		if (BASEPRI(ps)) {
			/*
			 * Save the overhead of a software interrupt;
			 * it will happen as soon as we return, so do it now.
			 */
			splsoftclock();
			softclock(pc, ps);
		} else
			setsoftclock();
#endif
	}
}

/*
 * Gather statistics on resource utilization.
 *
 * We make a gross assumption: that the system has been in the
 * state it is in (user state, kernel state, interrupt state,
 * or idle state) for the entire last time interval, and
 * update statistics accordingly.
 */
/*ARGSUSED*/
gatherstats(pc, ps)
	caddr_t pc;
	int ps;
{
	register int s;

	/*
	 * If the profiler is enabled take a snapshot
	 */
	 if (prfstat & PRF_ON)
		prfintr(pc, ps);

#ifdef HCXRT
#ifdef GPROF
	if (! USERMODE(ps)) {
		s = pc - s_lowpc;
		if (profiling < 2 && s < s_textsize)
			kcount[s / (HISTFRACTION * sizeof (*kcount))]++;
	}
#endif
#else
	/*
	 * We maintain statistics shown by user-level statistics
	 * programs:  the amount of time in each cpu state, and
	 * the amount of time each of DK_NDRIVE ``drives'' is busy.
	 */
	if (USERMODE(ps)) {
		/*
		 * CPU was in user state.
		 */
		if (u.u_procp->p_nice > NZERO)
			cp_time[CP_NICE]++;
		else
			cp_time[CP_USER]++;
	} else {
		/*
		 * CPU was in system state.  If profiling kernel
		 * increment a counter.  If no process is running
		 * then this is a system tick if we were running
		 * at a non-zero IPL (in a driver).  If a process is running,
		 * then we charge it with system time even if we were
		 * at a non-zero IPL, since the system often runs
		 * this way during processing of system calls.
		 * This is approximate, but the lack of true interval
		 * timers makes doing anything else difficult.
		 */
		if (noproc && BASEPRI(ps))
			cp_time[CP_IDLE]++;
		else
			cp_time[CP_SYS]++;
#ifdef GPROF
		s = pc - s_lowpc;
		if (profiling < 2 && s < s_textsize)
			kcount[s / (HISTFRACTION * sizeof (*kcount))]++;
#endif
	}

	for (s = 0; s < DK_NDRIVE; s++)
		if (dk_busy&(1<<s))
			dk_time[s]++;
#endif
}

/*
 * Software priority level clock interrupt.
 * Run periodic events from timeout queue.
 */
#ifdef HCXRT
softclock()
#else
/*ARGSUSED*/
softclock(pc, ps)
	caddr_t pc;
	int ps;
#endif
{

	for (;;) {
		register struct callout *p1;
		register caddr_t arg;
		register int (*func)();
		register int a, s;

		s = spl8();
		if ((p1 = calltodo.c_next) == 0 || p1->c_time > 0) {
			splx(s);
#ifdef HCXRT
			slumber((caddr_t)&proc[3], 0, noboost);
			continue;
#else
			break;
#endif
		}
		arg = p1->c_arg; func = p1->c_func; a = p1->c_time;
		calltodo.c_next = p1->c_next;
		p1->c_next = callfree;
		callfree = p1;
		splx(s);
		(*func)(arg, a);
	}
}

/*
 * Arrange that (*fun)(arg) is called in../sysz seconds.
 */
timeout(fun, arg, t)
	int (*fun)();
	caddr_t arg;
	register int t;
{
	register struct callout *p1, *p2, *pnew;
	register int s = spl8();

	if (t <= 0)
		t = 1;
	pnew = callfree;
	if (pnew == NULL)
		panic("timeout table overflow");
	callfree = pnew->c_next;
	pnew->c_arg = arg;
	pnew->c_func = fun;
	for (p1 = &calltodo; (p2 = p1->c_next) && p2->c_time < t; p1 = p2)
		if (p2->c_time > 0)
			t -= p2->c_time;
	p1->c_next = pnew;
	pnew->c_next = p2;
	pnew->c_time = t;
	if (p2)
		p2->c_time -= t;
	splx(s);
}

/*
 * untimeout is called to remove a function timeout call
 * from the callout structure.
 */
untimeout(fun, arg)
	int (*fun)();
	caddr_t arg;
{
	register struct callout *p1, *p2;
	register int s;

	s = spl8();
	for (p1 = &calltodo; (p2 = p1->c_next) != 0; p1 = p2) {
		if (p2->c_func == fun && p2->c_arg == arg) {
			if (p2->c_next && p2->c_time > 0)
				p2->c_next->c_time += p2->c_time;
			p1->c_next = p2->c_next;
			p2->c_next = callfree;
			callfree = p2;
			break;
		}
	}
	splx(s);
}


/*
 * Compute number of hz until specified time.
 * Used to compute third argument to timeout() from an
 * absolute time.
 */
hzto(tv)
	struct timeval *tv;
{
	register long ticks;
	register long sec;
	register long usec;
	int s = spl8();

	/*
	 * Maximum value for any timeout is about 415 days.
	 *
	 * If overflow, do the maximum wait allowed.
	 */
	sec = tv->tv_sec - time.tv_sec;
	usec = tv->tv_usec - time.tv_usec;
	if ( (0x7fffffff / hz - 1) > sec )
		ticks = sec * hz + usec / tick ;
	else
		ticks = 0x7fffffff;
	splx(s);
	return (ticks);
}


profil()
{
	register struct a {
		short	*bufbase;
		unsigned bufsize;
		unsigned pcoffset;
		unsigned pcscale;
	} *uap = (struct a *)u.u_ap;
#ifdef HCXRT
	u.u_error = EACCES;
#else
	register struct uprof *upp = &u.u_prof;

	upp->pr_base = uap->bufbase;
	upp->pr_size = uap->bufsize;
	upp->pr_off = uap->pcoffset;
	upp->pr_scale = uap->pcscale;
	/*
	 * Turn off profiling if an odd address was specified,
	 */
	if ( ((unsigned)(upp->pr_base)) & 1 ) {
		upp->pr_scale=0;
	}
#endif
}

delay(ticks)
	int ticks;
{
	extern wakeup();
	int	s = spl8();	/* tep 6/11/87 */

	if (ticks<=0)
		return;
	timeout(wakeup, (caddr_t)u.u_procp+1, ticks);
	slumber((caddr_t)u.u_procp+1, 0, delayboost);
	splx (s);
}

pause()
{

	for (;;)
		slumber((caddr_t)&u, INTERRUPTABLE, delayboost);
}

/*
 * Initialze the time of day register, based on the time base which is, e.g.
 * from a filesystem.
 */
inittodr(base)
	time_t base;
{
	time_t tbase;
	int     status;

	if (base < 5*SECYR)
	  {
	        printf("WARNING: preposterous time in file system");
	        base = 6*SECYR + 186*SECDAY + SECDAY/2;
	  }

	if ((status = readtocc(&tbase)) == 0)
	  {
	        if ((tbase - base < 0) || (tbase - base > SECDAY/12))
	          	printf("WARNING: file system time not current");
	        base = tbase;
	  }
	else if (status == 2) printf("WARNING: low tocc battery");
	time.tv_sec = base;
	printf(" CHECK AND RESET THE DATE!\n");

}
