/*	@(#)vfs_lookup.c 1.1 86/09/25 SMI	*/

#include "../sys/param.h"
#include "../sys/user.h"
#include "../sys/uio.h"
#include "../sys/vfs.h"
#include "../sys/vnode.h"
#include "../sys/dir.h"
#include "../sys/pathname.h"
#include "../sys/kernel.h"

/*
 * lookup the user file name,
 * Handle allocation and freeing of pathname buffer, return error.
 */
lookupname(fnamep, seg, followlink, dirvpp, compvpp)
	char *fnamep;			/* user pathname */
	int seg;			/* addr space that name is in */
	enum symfollow followlink;	/* follow sym links */
	struct vnode **dirvpp;		/* ret for ptr to parent dir vnode */
	struct vnode **compvpp;		/* ret for ptr to component vnode */
{
	struct pathname lookpn;
	register int error;

	error = lookuppathname(fnamep, seg, &lookpn, followlink,
		dirvpp, compvpp);
	pn_free(&lookpn);
	return (error);
}

/*
 * Starting at current directory, translate pathname pnp to end.
 *
 * This is the central routine in pathname translation and handles
 * multiple components in pathnames, separating them at /'s.  It also
 * implements mounted file systems and processes symbolic links.
 *
 * On return:
 *      pnp->pn_path, pnp->pn_len is final component of pathname.
 *      pnp->pn_buf is entire path name.
 *      *compvpp is the vnode of the final component.
 *	*dirvpp is the vnode of the parent of the final component.
 */
lookuppathname(fnamep, seg, pnp, followlink, dirvpp, compvpp)
	char *fnamep;			/* user pathname */
	int seg;			/* addr space that name is in */
	register struct pathname *pnp;	/* pathaname to lookup */
	enum symfollow followlink;	/* (don't) follow sym links */
	struct vnode **dirvpp;		/* ptr for parent vnode */
	struct vnode **compvpp;		/* ptr for entry vnode */
{
	register struct vnode *vp;	/* current directory vp */
	register struct vnode *cvp;	/* current component vp */
	struct vnode *tvp;		/* non-reg temp ptr */
	register struct vfs *vfsp;	/* ptr to vfs for mount indir */
	char component[MAXNAMLEN+1];	/* buffer for component */
	register int error;
	register int nlink;
	struct pathname component_pn ;

	error = pn_get( fnamep, seg, pnp ) ;
	if (error) {
		if (dirvpp != (struct vnode **)0) {
			*dirvpp = (struct vnode *)0;
		}
		if (compvpp != (struct vnode **)0) {
			*compvpp = (struct vnode *)0;
		}
	        return(error);
	}

	nlink = 0;
	cvp = (struct vnode *)0;

	/*
	 * start at current directory.
	 */
	vp = u.u_cdir;
	pnp->pn_type = CDIRREL ;
	VN_HOLD(vp);

begin:
	/*
	 * Each time we begin a new name interpretation (e.g.
	 * when first called and after each symbolic link is
	 * substituted), we allow the search to start at the
	 * root directory if the name starts with a '/', otherwise
	 * continuing from the current directory.
	 */
	component[0] = 0;
	if (pn_peekchar(pnp) == '/') {
		pnp->pn_type = ROOTREL ;
		VN_RELE(vp);
		pn_skipslash(pnp);
		if (u.u_rdir)
			vp = u.u_rdir;
		else
			vp = rootdir;
		VN_HOLD(vp);
	}

next:
	/*
	 * Make sure we have a directory.
	 */
	if (vp->v_type != VDIR) {
		error = ENOTDIR;
		goto bad;
	}
	/*
	 * Process the next component of the pathname.
	 */
	component_pn = *pnp ;
	error = pn_getcomponent(pnp, component);
	if (error)
		goto bad;

	/*
	 * Check for degenerate name (e.g. / or "")
	 * which is a way of talking about a directory,
	 * e.g. "/." or ".".
	 */
	if (component[0] == 0) {
		/*
		 * If the caller was interested in the parent then
		 * return an error since we don't have the real parent
		 */
		if (dirvpp != (struct vnode **)0) {
			VN_RELE(vp);
			return(EINVAL);
		}
		pn_set(pnp, ".");
		if (compvpp != (struct vnode **)0) {
			*compvpp = vp;
		} else {
			VN_RELE(vp);
		}
		return(0);
	}

	/*
	 * Handle "..": two special cases.
	 * 1. If at root directory (e.g. after chroot)
	 *    then ignore it so can't get out.
	 * 2. If this vnode is the root of a mounted
	 *    file system, then replace it with the
	 *    vnode which was mounted on so we take the
	 *    .. in the other file system.
	 */
	if (strcmp(component, "..") == 0) {
checkforroot:
#ifdef notdef
		printf("vfs_lookup: .. found\n");
#endif
		if ((vp == u.u_rdir) || (vp == rootdir)) {
			cvp = vp;
			VN_HOLD(cvp);
			goto skip;
		}
		if (vp->v_flag & VROOT) {
			cvp = vp;
			vp = vp->v_vfsp->vfs_vnodecovered;
#ifdef notdef
			printf("vfs_lookup: VROOT cvp=%d vp=%d\n",cvp,vp);
#endif
			VN_HOLD(vp);
			VN_RELE(cvp);
			cvp = (struct vnode *)0;
			goto checkforroot;
		}
	}

	/*
	 * Perform a lookup in the current directory.
	 */
	error = VOP_LOOKUP(vp, component, &tvp, u.u_cred);
#ifdef notdef
	printf( "vfs_lookup: vop_lookup: %s cvp=%d vp=%d\n", component,tvp,vp);
#endif
	cvp = tvp;
	if (error) {
		cvp = (struct vnode *)0;
		/*
		 * On error, if more pathname or if caller was not interested
		 * in the parent directory then hard error.
		 * If the path is unreadable, fail now with the right error.
		 */
		if (pn_pathleft(pnp) || dirvpp == (struct vnode **)0 ||
		    (error == EACCES))
			goto bad;
		*pnp=component_pn;
		*dirvpp = vp;
		if (compvpp != (struct vnode **)0) {
			*compvpp = (struct vnode *)0;
		}
		return (0);
	}
	/*
	 * If we hit a symbolic link and there is more path to be
	 * translated or this operation does not wish to apply
	 * to a link, then place the contents of the link at the
	 * front of the remaining pathname.
	 */
	if (cvp->v_type == VLNK &&
	    ((followlink == FOLLOW_LINK) || pn_pathleft(pnp))) {
		struct pathname linkpath;

		nlink++;
		if (nlink > MAXSYMLINKS) {
			error = ELOOP;
			goto bad;
		}
		error = getsymlink(cvp, &linkpath);
		if (error)
			goto bad;
		if (pn_pathleft(&linkpath) == 0)
			pn_set(&linkpath, ".");
		error = pn_combine(pnp, &linkpath);	/* linkpath before pn */
		pn_free(&linkpath);
		if (error)
			goto bad;
		VN_RELE(cvp);
		cvp = (struct vnode *)0;
		goto begin;
	}

	/*
	 * If this vnode is mounted on, then we
	 * transparently indirect to the vnode which 
	 * is the root of the mounted file system.
	 * Before we do this we must check that an unmount is not
	 * in progress on this vnode. This maintains the fs status
	 * quo while a possibly lengthy unmount is going on.
	 */
mloop:
	while (vfsp = cvp->v_vfsmountedhere) {
		while (vfsp->vfs_flag & VFS_MLOCK) {
			vfsp->vfs_flag |= VFS_MWAIT;
			slumber((caddr_t)vfsp, INTERRUPTABLE, iocomboost);
			goto mloop;
		}
		error = VFS_ROOT(cvp->v_vfsmountedhere, &tvp);
		if (error)
			goto bad;
		VN_RELE(cvp);
		cvp = tvp;
	}

skip:
	/*
	 * Skip to next component of the pathname.
	 * If no more components, return last directory (if wanted)  and
	 * last component (if wanted).
	 */

	/* Skip trailing '/' on directories. */
	if (vp->v_type == VDIR)
		pn_skipslash(pnp);

	if (pn_pathleft(pnp) == 0) {
	        if (component_pn.pn_path[component_pn.pn_pathlen - 1] == '/') {
	        	component_pn.pn_path[component_pn.pn_pathlen -1 ] = '\0';
	        	(component_pn.pn_pathlen)--;
	        }
		*pnp=component_pn;
		if (dirvpp != (struct vnode **)0) {
			/*
			 * check that we have the real parent and not
			 * an alias of the last component
			 */
			if (vp == cvp) {
				VN_RELE(vp);
				VN_RELE(cvp);
				return(EINVAL);
			}
			*dirvpp = vp;
		} else {
			VN_RELE(vp);
		}
		if (compvpp != (struct vnode **)0) {
			*compvpp = cvp;
		} else {
			VN_RELE(cvp);
		}
		return (0);
	}
	/*
	 * skip over slashes from end of last component
	 */
	pn_skipslash(pnp);

	/*
	 * Searched through another level of directory:
	 * release previous directory handle and save new (result
	 * of lookup) as current directory.
	 */
	VN_RELE(vp);
	vp = cvp;
	cvp = (struct vnode *)0;
	goto next;

bad:
	/*
	 * Error. Release vnodes and return.
	 */
	if (cvp)
		VN_RELE(cvp);
	VN_RELE(vp);
	return (error);
}

/*
 * Gets symbolic link into pathname.
 */
static int
getsymlink(vp, pnp)
	struct vnode *vp;
	register struct pathname *pnp;
{
	struct iovec aiov;
	struct uio auio;
	register int i ;
	register char * cp ;
	register int error;

	pn_alloc(pnp);
	aiov.iov_base = pnp->pn_buf;
	aiov.iov_len = MAXPATHLEN;
	auio.uio_iov = &aiov;
	auio.uio_iovcnt = 1;
	auio.uio_offset = 0;
	auio.uio_segflg = UIO_SYSSPACE;
	auio.uio_resid = MAXPATHLEN;
	error = VOP_READLINK(vp, &auio, u.u_cred);
	pnp->pn_pathlen = MAXPATHLEN - auio.uio_resid;
	if (error) {
		pn_free(pnp);
		return( error ) ;
	}

	/*
	 * Detect and replace any references to universe name.
	 */

	for ( i=0 , cp=pnp->pn_path ; i < pnp->pn_pathlen ; i ++ , cp++ ) {
		if ( *cp == '{' && i+1 < pnp->pn_pathlen && *(cp+1) == '}' ) {
		        if ( pnp->pn_pathlen + u.u_univlen - 3 > MAXPATHLEN ) {
		                pn_free( pnp ) ;
		                return( ENAMETOOLONG ) ;
		        }
		        ovbcopy( cp+2, cp+u.u_univlen-1, pnp->pn_pathlen - i-2);
		        bcopy( u.u_univ, cp, u.u_univlen-1 ) ;
		        i += u.u_univlen - 2 ;
		        cp += u.u_univlen - 2 ;
		        pnp->pn_pathlen += u.u_univlen - 3 ;
		}
	}
	return ( 0 );
}
